"""Module containing tracking metric computation"""

import networkx as nx
import numpy as np


def compute_positive_division(
    pred_lc_lineage: nx.DiGraph, gt_lc_lineage: nx.DiGraph
) -> tuple:
    """Computes the counts of true positive (tp) and false positive (fp) cell divisions

    Args:
        pred_lc_lineage (nx.DiGraph): predicted life cycle lineage (every cell track is one node)
        gt_lc_lineage (nx.DiGraph): ground truth life cycle lineage (every cell track is one node)

    Returns:
        tuple: true-positive (tp) and false positive (fp) counts
    """

    tp = 0
    fp = 0

    # create cycle lookup dict
    gt_lc_lookup = {}
    for gt_lc_node in gt_lc_lineage:
        for gt_node in gt_lc_lineage.nodes[gt_lc_node]["cycle"]:
            gt_lc_lookup[gt_node] = gt_lc_node

    for pred_node in pred_lc_lineage.nodes:
        pred_successors = list(pred_lc_lineage.successors(pred_node))
        pred_cycle = pred_lc_lineage.nodes[pred_node]["cycle"]

        if len(pred_successors) > 1:
            # this is a division

            # identify the respective gt node by using the last
            gt_node = gt_lc_lookup[pred_cycle[-1]]
            gt_cycle = gt_lc_lineage.nodes[gt_node]["cycle"]
            gt_successors = list(gt_lc_lineage.successors(gt_node))

            # direct parent (last cell detection in the cycle) must be the same
            match = True

            match &= pred_cycle[-1] == gt_cycle[-1]

            # direct children (first cell detection in the cycle) must be the same
            pred_direct_children = {
                pred_lc_lineage.nodes[succ]["cycle"][0] for succ in pred_successors
            }
            gt_direct_children = {
                gt_lc_lineage.nodes[succ]["cycle"][0] for succ in gt_successors
            }

            match &= pred_direct_children == gt_direct_children

            if match:
                # same division exists in gt
                tp += 1
            else:
                # division does not exist in gt
                fp += 1

    return tp, fp


def compute_bc_stats(pred_lc_lineage: nx.DiGraph, gt_lc_lineage: nx.DiGraph) -> dict:
    """Compute the branch correctness (bc) statistics

    Args:
        pred_lc_lineage (nx.DiGraph): predicted life cycle lineage (every cell track is one node)
        gt_lc_lineage (nx.DiGraph): ground truth life cycle lineage (every cell track is one node)

    Returns:
        dict: dictionary with true positive (tp), false positive (fp) and false negative (fn) counts
    """

    tp = 0
    fp = 0
    fn = 0

    # compare pred divisions with gt divisions
    tp, fp = compute_positive_division(
        pred_lc_lineage=pred_lc_lineage, gt_lc_lineage=gt_lc_lineage
    )
    # do it the other way round
    tp2, fn = compute_positive_division(
        pred_lc_lineage=gt_lc_lineage, gt_lc_lineage=pred_lc_lineage
    )

    # make sure that tps agree
    np.testing.assert_equal(tp, tp2)

    return dict(tp=tp, fp=fp, fn=fn)


def compute_positive_cts(pred_lc_lineage, gt_lc_lineage) -> tuple:
    """Computes the counts of true positive (tp) and false positive (fp) cell tracks (cts)

    Args:
        pred_lc_lineage (nx.DiGraph): predicted life cycle lineage (every cell track is one node)
        gt_lc_lineage (nx.DiGraph): ground truth life cycle lineage (every cell track is one node)

    Returns:
        tuple: tp and fp counts
    """
    tp = 0
    fp = 0

    gt_lc_lookup = {}
    for gt_lc_node in gt_lc_lineage:
        for gt_node in gt_lc_lineage.nodes[gt_lc_node]["cycle"]:
            gt_lc_lookup[gt_node] = gt_lc_node

    for cell_cycle_node in pred_lc_lineage.nodes:
        pred_cc = list(pred_lc_lineage.nodes[cell_cycle_node]["cycle"])

        match = False

        # get the gt cell cycle by the first node
        gt_cc_node = gt_lc_lookup[pred_cc[0]]
        gt_cc = list(gt_lc_lineage.nodes[gt_cc_node]["cycle"])

        if pred_cc == gt_cc:
            # they cover the same cell track

            match = True

        if match is False:
            # no matching cell track was found -> it is a false positive
            fp += 1
        else:
            # matching cell track was found!
            tp += 1

    return tp, fp


def compute_ct_stats(pred_lc_lineage, gt_lc_lineage):
    """Compute the correct cell track (bc) statistics. A cell track is correct if from cell birth to cell division/death/disappearance all edges are correctly reconstructed. It does not deal with division edges.

    Args:
        pred_lc_lineage (nx.DiGraph): predicted life cycle lineage (every cell track is one node)
        gt_lc_lineage (nx.DiGraph): ground truth life cycle lineage (every cell track is one node)

    Returns:
        _type_: _description_
    """
    tp = 0
    fp = 0
    fn = 0

    # compare gt tracks with predicition
    tp, fp = compute_positive_cts(
        pred_lc_lineage=pred_lc_lineage, gt_lc_lineage=gt_lc_lineage
    )
    # compare pred tracks with gt
    tp2, fn = compute_positive_cts(
        pred_lc_lineage=gt_lc_lineage, gt_lc_lineage=pred_lc_lineage
    )

    # make sure the tps agree
    np.testing.assert_equal(tp, tp2)

    return dict(tp=tp, fp=fp, fn=fn)


def compute_f1(tp: int, fp: int, fn: int) -> dict:
    """Compute f1 score from count statistics

    Args:
        tp (int): true positive (tp) count
        fp (int): false positive (fp) count
        fn (int): false negative (fn) count

    Returns:
        dict: dictionary with precision, recall and f1 scores
    """
    precision = tp / (tp + fp)
    recall = tp / (tp + fn)

    f1 = 2 / (1 / precision + 1 / recall)

    return dict(precision=precision, recall=recall, f1=f1)



""" Module for tracking result output generation """

from __future__ import annotations

import gzip
import json
import time
from pathlib import Path

from pandas import DataFrame

from uatrack.core import SimpleCluster


class SimpleTrackingReporter:
    """Generate output in the simple tracking format"""

    def __init__(
        self, output_folder: str, df: DataFrame, detections, assignment_generators
    ):
        self.all_clusters = []
        self.output_folder = output_folder
        self.df = df
        self.detections = detections
        self.timings = []
        self.timings.append(time.time())
        self.assignment_generators = assignment_generators

        self.final_cluster = None

    def report_distribution(self, current_clusters: list[SimpleCluster]):
        self.final_cluster = current_clusters[0]

    def close(self):
        Path(self.output_folder).mkdir(exist_ok=True)
        edge_list = list(
            map(
                lambda e: (int(e[0]), int(e[1])),
                self.final_cluster.tracking.createIndexTracking().edges,
            )
        )

        tracking_data = [
            dict(
                sourceId=self.detections[edge[0]].id,
                targetId=self.detections[edge[1]].id,
            )
            for edge in edge_list
        ]
        segmentation_data = [
            dict(
                label=cont.label,
                contour=cont.coordinates.tolist(),
                id=cont.id,
                frame=cont.frame,
            )
            for cont in self.detections
        ]

        data_structure = dict(
            segmentation=segmentation_data,
            tracking=tracking_data,
            format_version="0.0.1",
        )

        print(self.output_folder)

        with gzip.open(
            Path(self.output_folder) / "tracking.json.gz", "wt"
        ) as output_file:
            json.dump(data_structure, output_file)
        # print(self.final_cluster.tracking.createIndexTracking().edges)



""" Module for simple sampling functionality"""

import numbers

import numpy as np


class Probability:
    """Probability value that uses log probability for computations"""

    def __init__(self, probability=None, log_prob=None):
        assert probability is None or log_prob is None

        if probability is not None:
            assert np.all(probability > 0)
            self.logProb = np.log(probability)
        else:
            self.logProb = log_prob

    @property
    def probability(self):
        return np.exp(self.logProb)

    @property
    def log_probability(self):
        return self.logProb

    def __eq__(self, other):
        if isinstance(other, numbers.Number):
            return self.probability == other
        elif isinstance(other, Probability):
            return self.logProb == other.logProb
        else:
            # unsupported equals operator
            assert False

    def __pow__(self, other):
        assert isinstance(other, numbers.Number)
        return Probability(log_prob=self.logProb * other)

    def __mul__(self, other):
        temp = other
        if not isinstance(other, Probability):
            temp = Probability(other)
        return Probability(log_prob=self.logProb + temp.logProb)

    def __rmul__(self, other):
        """
        other: some kind of number (int, float) that is used on left side
        """
        return self.__mul__(Probability(other))

    def __truediv__(self, other):
        temp = other
        if not isinstance(other, Probability):
            temp = Probability(other)
        return Probability(log_prob=self.logProb - temp.logProb)

    def __rtruediv__(self, other):
        temp = other
        if not isinstance(other, Probability):
            temp = Probability(other)
        return Probability(log_prob=temp.logProb - self.logProb)

    def __add__(self, other):
        # According to https://en.wikipedia.org/wiki/Log_probability
        # log(x + y) = x' + log(1 + exp(y' - x')) where x'=log(x) and y'=log(y)
        temp = other
        if not isinstance(other, Probability):
            temp = Probability(other)
        return Probability(
            log_prob=self.logProb + np.log1p(np.exp(temp.logProb - self.logProb))
        )

    def __radd__(self, other):
        if other == 0:
            # Special case: if we add 0 to probability it stays unchanged!
            return Probability(log_prob=self.logProb)
        return self.__add__(Probability(other))

    def __sub__(self, other):
        # According to https://en.wikipedia.org/wiki/Log_probability
        # log(x + y) = x' + log(1 + exp(y' - x')) where x'=log(x) and y'=log(y)
        # therefore log(x - y) = x' + log(1 - exp(y' - x'))
        temp = other
        if not isinstance(other, Probability):
            temp = Probability(other)

        assert np.all(self.log_probability >= temp.log_probability)

        return Probability(
            log_prob=self.logProb + np.log1p(-np.exp(temp.logProb - self.logProb))
        )

    def __rsub__(self, other):
        if isinstance(other, float):
            assert other > 0 and np.all(np.log(other) >= self.log_probability)

            other = Probability(other)

        return other.__sub__(self)

    def __str__(self):
        return f"Probability(value: {self.probability:.4f}, log: {self.logProb:.2f})"


def log_prob_dist_to_prob_dist(raw_log_probabilities):
    """
    Takes a list of log probablities, normalizes them and converts them to probability values

    raw_log_probabilities: list of log probabilities
    """
    # convert to our probability space
    probabilites = [
        Probability(log_prob=log_prob) for log_prob in raw_log_probabilities
    ]
    # compute discrete distribution normalizer
    normalizer = sum(probabilites)

    return [(prob / normalizer).probability for prob in probabilites]


def sample_from_log_probabilities(raw_log_probabilities, num_samples):
    """
    Samples indices from a raw log probability list
    """
    prob_dist = log_prob_dist_to_prob_dist(raw_log_probabilities)

    return np.random.choice(
        range(len(raw_log_probabilities)), size=num_samples, p=prob_dist
    )


def normalize_prob_dist(probabilities):
    normalizer = sum(probabilities)
    return [prob / normalizer for prob in probabilities]


def sample_from_probabilities(probabilities, num_samples):
    normed_dist = normalize_prob_dist(probabilities)
    value_probabilities = [prob.probability for prob in normed_dist]

    return np.random.choice(
        range(len(probabilities)), size=num_samples, p=value_probabilities
    )


if __name__ == "__main__":
    # small test
    p = Probability(np.linspace(0.1, 0.7, 7))
    p2 = Probability(np.linspace(0.2, 0.8, 7, dtype=np.float32))

    print(p.log_probability.shape, p2.log_probability.shape)

    newProb = 1.0 - (p2 - p)

    print(newProb.probability)



"""Testing assignment generation"""

import unittest

import numpy as np

from uatrack.assignment import SimpleNewAssGenerator, filter_targets


class TestAssignmentGenerators(unittest.TestCase):
    """Test contour functionality"""

    def test_new(self):
        models = []

        ass_gen = SimpleNewAssGenerator(models)

        targets = np.arange(0, 10, 1, dtype=np.int32)

        # test that empty model leads to problems
        with self.assertRaises(ValueError):
            ass_gen.generate(None, None, targets)

        def simple_score(tr, s, t):
            # pylint: disable=unused-argument
            return np.zeros((len(t),), dtype=np.float32)

        models = [simple_score]

        ass_gen = SimpleNewAssGenerator(models)

        assignments = ass_gen.generate(None, None, targets)

        self.assertEqual(len(assignments[0]), 10)

    def test_filter(self):

        # dummy example to filter out some assignments
        sources = np.arange(0, 10, 1, dtype=np.uint32)
        targets = np.copy(sources)

        source_index = sources
        target_index = target_index = np.tile(np.array(targets), (len(sources), 1))

        # pylint: disable=unused-argument
        def filter(a, b):
            return b < 5

        filters = [filter]

        new_source_index, new_target_index = filter_targets(
            source_index, target_index, filters=filters
        )

        self.assertEqual(len(new_source_index), 10)
        self.assertEqual(new_target_index.size, 50)


if __name__ == "__main__":
    unittest.main()



""" This module contains the core functionality """

from __future__ import annotations

import copy
import json
import logging
import multiprocessing
import time
from functools import reduce
from pathlib import Path

import numpy as np
import numpy.typing as npt
import ray
from networkx import DiGraph
from pandas import DataFrame
from tqdm.auto import tqdm

from uatrack.sampling import sample_from_probabilities
from uatrack.solve.gurobi import SimpleGurobiSolver
from uatrack.solve.mip import SimpleMIPSolver

from .sampling import Probability
from .utils import ContourDistanceCache


class SimpleTracking:
    """Contains a tracking solution in simple array formats"""

    def __init__(self, tracking, parent_id=-1, id=-1, active_assignments=None):
        self.id = id
        self.parent_id = parent_id
        self.parents = tracking

        if active_assignments is None:
            active_assignments = []

        self.active_assignments = active_assignments

    def compute_trackends(self, lastFrameIndices):
        """
        computes the current active trackends

        lastFrameIndices: all indices the are candidates for being track ends
        """

        # trackends are those that have a parent, or roots (-1)
        return lastFrameIndices[self.parents[lastFrameIndices] >= -1]

    def expand(self, solution, assignments):
        """
        Expend the current tracking with a new solution

        solution: solution structure (mask, score)
        assignments: corresponding assignment structure
        """
        mask, _ = solution
        assignments = zip(*assignments)

        new_parents = np.copy(self.parents)
        start_index = 0

        active_assignments = []

        # iterate over different assignment types
        for ass_type in assignments:
            sources, targets, scores, ind_scores = ass_type

            num_assignments = len(scores)

            # compute selection mask
            end_index = start_index + num_assignments
            sel_mask = mask[start_index:end_index]

            # repeat the sources of active assignments
            rep_sources = np.repeat(sources[sel_mask].flatten(), targets.shape[1])

            # update the new parent array
            new_parents[targets[sel_mask].flatten()] = rep_sources

            # move selection mask indices
            start_index += num_assignments

            active_assignments += [
                (
                    sources[sel_mask],
                    targets[sel_mask],
                    scores[sel_mask],
                    ind_scores[sel_mask],
                )
            ]

        assert start_index == len(mask)

        # return the new tracking
        return SimpleTracking(
            new_parents,
            active_assignments=self.active_assignments + [active_assignments],
            parent_id=self.id,
        )

    def createIndexTracking(self):
        graph = DiGraph()

        for i, p in enumerate(self.parents):
            if p >= 0:
                graph.add_edge(p, i)

        return graph

    def createGraphTracking(self, detections):
        """
        convert the simple tracking to networkx graph containing the detections as node
        """
        graph = DiGraph()

        np_detections = np.array(detections, dtype=np.object)

        graph.add_nodes_from(np_detections[np.where(self.parents >= -1)])

        indices = np.arange(len(self.parents))
        mask = self.parents >= 0
        sel_indices = indices[mask]
        edges = np.stack([self.parents[sel_indices], sel_indices], axis=-1)

        graph.add_edges_from(np_detections[edges])

        return graph


class SimpleCluster:
    """
    Cluster groups several particles with the same tracking

    This allows more efficient inference
    """

    def __init__(
        self, tracking: SimpleTracking, size: int, cumulated_prob: Probability
    ):
        self.tracking = tracking
        self.size = size
        assert isinstance(cumulated_prob, Probability)
        self.cumulated_prob = cumulated_prob


@ray.remote
def simpleExpansionRay(
    cluster,
    sources,
    targets,
    assignment_generators,
    cutOff,
    max_num_solutions,
    pool_gap,
):
    return simpleExpansion(
        cluster,
        sources,
        targets,
        assignment_generators,
        cutOff,
        max_num_solutions,
        pool_gap=pool_gap,
    )


@ray.remote
def singleSolutionRay(
    cluster: SimpleCluster,
    sources: npt.ArrayLike[np.uint32],
    targets: npt.ArrayLike[np.uint32],
    assignment_generators: list,
    index=int,
):
    return index, singleSolution(cluster, sources, targets, assignment_generators)


def generateAssignments(
    cluster: SimpleCluster, sources, targets, assignment_generators: list
):
    """
    Generates and scores all the possible assignments from sources to targets

    sources: index list of possible sources
    targets: index list of possible targets
    assignment_generators: list of assignment generators
    """
    all_assignments = []

    # compute all possible assignments and compute their costs
    for ass_gen in assignment_generators:
        source_index, target_index, scores, individual_scores = ass_gen.generate(
            cluster.tracking, sources, targets
        )
        all_assignments.append((source_index, target_index, scores, individual_scores))

    all_assignments = list(zip(*all_assignments))

    return all_assignments


def buildMIPProblem(all_assignments, cutOff=None, solver_name="GRB"):
    """
    builds the integer program from set of assignments

    all_assignments: list of different types of assignments
    cutOff: log value of solution search cutOff
    """

    if solver_name == "GRB":
        # build MIP solver using gurobipy
        solver = SimpleGurobiSolver(all_assignments)
    else:
        # build mip solver using mip (pypi library)
        solver = SimpleMIPSolver(all_assignments, solver_name=solver_name)

    if cutOff:
        solver.setCutOff(cutOff)

    return solver


def singleSolution(
    cluster: SimpleCluster,
    sources,
    targets,
    assignment_generators: list,
    mip_method="auto",
):
    """
    Computes the best solution for an expansion

    cluster: the cluster containing the tracking until now
    sources: source index list
    targets: target index list
    assignment_generators: list of assignment generators
    """

    # generate assignments
    all_assignments = generateAssignments(
        cluster, sources, targets, assignment_generators
    )

    # build integer linear problem
    solver = buildMIPProblem(all_assignments, solver_name=mip_method)

    # solve for single solution
    max_sol = solver.solve()

    return None, max_sol


def simpleExpansion(
    cluster: SimpleCluster,
    sources,
    targets,
    assignment_generators: list,
    cutOff=None,
    max_num_solutions=50,
    pool_gap=None,
    mip_method="auto",
) -> list[tuple[SimpleCluster, Probability]]:
    """
    Computes an expansion step for a cluster

    cluster: cluster to use
    sources: df restricted table containing only source detections
    targets: df restricted table containing only target detections
    assignment_generators: list of assignment generators
    """
    # compute all possible assignments and compute their costs
    start = time.time()
    all_assignments = generateAssignments(
        cluster, sources, targets, assignment_generators
    )
    end = time.time()
    ass_gen_dur = end - start

    # compute cheapest tracking solutions with gurobi
    start = time.time()
    solver = buildMIPProblem(all_assignments, cutOff, solver_name=mip_method)
    all_solutions = solver.populate(
        max_num_solutions=max_num_solutions, pool_gap=pool_gap
    )
    end = time.time()

    # print('Num solutions', len(all_solutions))

    opt_solve_dur = end - start

    # print('ass_gen', ass_gen_dur, 'opt_solve', opt_solve_dur)

    # this is our truncated distribution proposal distribution from (9)
    sol_probabilities = [Probability(log_prob=sol[1]) for sol in all_solutions]

    if len(sol_probabilities) == 0:
        logging.warning("No expansions found")
        return []

    sampled_sol_indices = sample_from_probabilities(sol_probabilities, cluster.size)

    # According to (8) from UAT we have to sum over the full distribution
    weight = sum(
        sol_probabilities
    )  # this weight is for a single particle, multiply by the size of the cluster

    # group them to clusters and append to a new cluster list
    unique, counts = np.unique(sampled_sol_indices, return_counts=True)
    cluster_expansions = zip(unique, counts)
    new_cluster_candidates: list[tuple[SimpleCluster, float]] = []
    for solution_index, size in cluster_expansions:
        c = SimpleCluster(
            cluster.tracking.expand(all_solutions[solution_index], all_assignments),
            size,
            cumulated_prob=cluster.cumulated_prob * sol_probabilities[solution_index],
        )
        new_cluster_candidates.append((c, weight * size))

    logging.info(
        f"Expansion Step {ass_gen_dur + opt_solve_dur:.2f}s (assgen: {ass_gen_dur:.2f}s, solve: {opt_solve_dur:.2f}s)"
    )

    return new_cluster_candidates


def computeBestExpansions(
    current_cluster_dist,
    sources,
    targets,
    assignment_generators,
    num_cores,
    report_progress,
    use_ray,
    mip_method="auto",
):
    single_solutions = []

    # decide parallelized execution
    if num_cores == 1:
        # iterate over clusters and expand one by one
        it = current_cluster_dist

        if report_progress:
            it = tqdm(current_cluster_dist, leave=False)

        for cluster in it:
            single_solutions.append(
                singleSolution(
                    cluster,
                    sources,
                    targets,
                    assignment_generators,
                    mip_method=mip_method,
                )
            )
    elif use_ray:
        # prepare ray
        sources_remote = ray.put(sources)
        targets_reomte = ray.put(targets)
        assignment_generators_reomte = ray.put(assignment_generators)
        # perform computation
        single_solutions = [
            singleSolutionRay.remote(
                cluster=c,
                sources=sources_remote,
                targets=targets_reomte,
                assignment_generators=assignment_generators_reomte,
                index=i,
            )
            for i, c in enumerate(current_cluster_dist)
        ]

        def to_iterator(obj_ids):
            while obj_ids:
                done, obj_ids = ray.wait(obj_ids)
                yield ray.get(done[0])

        ray_it = to_iterator(
            single_solutions
        )  # ray.get(new_cluster_candidates_and_weights)

        if report_progress:
            ray_it = tqdm(ray_it, total=len(single_solutions), leave=False)

        split_solutions = list(zip(*list(ray_it)))

        logging.info(split_solutions[0])
        order = np.argsort(np.array(list(split_solutions[0]), dtype=np.int32))

        # reorder single solutions so that they match the cluster list (this is necessary due to multiprocessing)
        single_solutions = np.array(list(split_solutions[1]), dtype=np.object)[
            order
        ]  # reduce(lambda a,b: a+b, ray_it)
    else:
        # use multiprocessing for parallelization
        with multiprocessing.Pool(num_cores) as p:
            single_solutions = list(
                p.map(
                    lambda c: singleSolution(
                        c, sources, targets, assignment_generators
                    ),
                    current_cluster_dist,
                    chunksize=4,
                )
            )

    # return
    return single_solutions


def computeAllExpansionsAdv(
    current_cluster_dist,
    sources,
    targets,
    assignment_generators,
    num_cores,
    report_progress,
    use_ray,
    truncation_threshold=np.log(1e-3),
    max_num_solutions=50,
    mip_method="auto",
):
    # compute all best expansions
    best_expansions = computeBestExpansions(
        current_cluster_dist,
        sources,
        targets,
        assignment_generators,
        num_cores,
        report_progress,
        use_ray,
        mip_method=mip_method,
    )

    # compute the maximum likely solution
    best_exps_zipped = list(zip(*best_expansions))
    maxSols = np.array(list(map(lambda sol: sol[1], best_exps_zipped[1])))

    # get maximum possible cumulated probability
    cumLogProbs = np.array(
        [c.cumulated_prob.log_probability for c in current_cluster_dist]
    )
    bestCumSol = np.max(maxSols + cumLogProbs)

    # best possible cumulated probability for every cluster (cluster cumProb * expansion prob)
    individualBestCumLogProbs = maxSols + cumLogProbs

    # compute truncation threshold based on that
    truncation = truncation_threshold
    cutOff = bestCumSol + truncation
    truncation_mask = individualBestCumLogProbs < cutOff
    pruned_clusters = np.sum(truncation_mask)

    logging.info(f"Possible pruned clusters: {pruned_clusters}")

    # for the solution pool we must compute the relative gap between the best solution and the worst solution: worst = best + gap * best
    individual_pool_gaps = (
        cutOff - individualBestCumLogProbs
    ) / individualBestCumLogProbs

    # pruning cluster distributions and pool gaps
    pruned_clusters_dist = np.array(current_cluster_dist)[~truncation_mask]
    individual_pool_gaps = individual_pool_gaps[~truncation_mask]

    logging.info(individual_pool_gaps)

    return computeAllExpansions(
        pruned_clusters_dist,
        sources,
        targets,
        assignment_generators,
        num_cores,
        report_progress,
        use_ray,
        cutOff=cutOff,
        max_num_solutions=max_num_solutions,
        individual_pool_gaps=individual_pool_gaps,
        mip_method=mip_method,
    )


def computeAllExpansions(
    current_cluster_dist,
    sources,
    targets,
    assignment_generators,
    num_cores,
    report_progress,
    use_ray,
    individual_pool_gaps,
    cutOff=None,
    max_num_solutions=50,
    mip_method="auto",
):
    new_cluster_candidates_and_weights = []

    # decide parallelized execution
    if num_cores == 1:
        # iterate over clusters and expand one by one
        it = current_cluster_dist

        if report_progress:
            it = tqdm(current_cluster_dist, leave=False)

        for i, cluster in enumerate(it):
            new_cluster_candidates_and_weights += simpleExpansion(
                cluster,
                sources,
                targets,
                assignment_generators,
                cutOff,
                max_num_solutions=max_num_solutions,
                pool_gap=individual_pool_gaps[i],
                mip_method=mip_method,
            )
    elif use_ray:
        # prepare ray
        sources_remote = ray.put(sources)
        targets_reomte = ray.put(targets)
        assignment_generators_reomte = ray.put(assignment_generators)
        # perform computation
        new_cluster_candidates_and_weights = [
            simpleExpansionRay.remote(
                cluster=c,
                sources=sources_remote,
                targets=targets_reomte,
                assignment_generators=assignment_generators_reomte,
                cutOff=cutOff,
                max_num_solutions=max_num_solutions,
                pool_gap=individual_pool_gaps[i],
            )
            for i, c in enumerate(current_cluster_dist)
        ]

        def to_iterator(obj_ids):
            while obj_ids:
                done, obj_ids = ray.wait(obj_ids)
                yield ray.get(done[0])

        ray_it = to_iterator(
            new_cluster_candidates_and_weights
        )  # ray.get(new_cluster_candidates_and_weights)

        if report_progress:
            ray_it = tqdm(
                ray_it, total=len(new_cluster_candidates_and_weights), leave=False
            )

        new_cluster_candidates_and_weights = reduce(lambda a, b: a + b, ray_it)
    else:
        # use multiprocessing for parallelization
        with multiprocessing.Pool(num_cores) as p:
            new_cluster_candidates_and_weights += reduce(
                lambda a, b: a + b,
                p.map(
                    lambda c: simpleExpansion(
                        c[1],
                        sources,
                        targets,
                        assignment_generators,
                        cutOff,
                        max_num_solutions=max_num_solutions,
                        pool_gap=individual_pool_gaps[c[0]],
                    ),
                    enumerate(current_cluster_dist),
                    chunksize=4,
                ),
                [],
            )

    return new_cluster_candidates_and_weights


def simpleTracking(
    df: DataFrame,
    assignment_generators,
    num_particles,
    num_cores,
    use_ray=True,
    reporters=None,
    report_progress=True,
    max_num_hypotheses=int(1e3),
    cutOff=np.log(1e-3),
    max_num_solutions=50,
    saving_interval=500,
    mip_method="auto",
    frames=None,
):
    """
    df: data frame containing all cell information
    assignment_generators: list of assignment generators
    num_particles: number of particles for the particle filter
    num_cores: number of execution cores to be used
    use_ray: if true and more than one core should be used, ray package is used for parallelization otherwise multiprocessing is used
    """

    if reporters is None:
        reporters = []

    # try to launch ray (only if really needed be)
    if num_cores > 1 and use_ray and not ray.is_initialized():
        ray.init(num_cpus=num_cores)

    # get the frame list
    if frames is None:
        frames = np.unique(df["frame"])

    logging.info(f"Frames: {frames}")

    # create initial empty cluster with full size and probability 1
    current_cluster_dist = [
        SimpleCluster(
            SimpleTracking(
                np.ones((len(df),), dtype=np.int32) * -2, parent_id=-1, id=0
            ),
            size=num_particles,
            cumulated_prob=Probability(1),
        )
    ]
    next_tracking_id = 1
    # all_clusters = [current_cluster_dist]

    try:
        # loop over consecutive frames
        for frame_index, (source_frame, target_frame) in enumerate(
            tqdm(
                zip(frames, frames[1:]), total=len(frames) - 1, desc="Perform tracking"
            )
        ):
            logging.info(f"Frames: {source_frame} -> {target_frame}")

            # find source and target detections
            sources = df[df["frame"] == source_frame].index.to_numpy(
                dtype=np.int32
            )  # TODO find only trackends
            targets = df[df["frame"] == target_frame].index.to_numpy(dtype=np.int32)

            logging.info(f"trackEnds {len(sources)} --> {(len(targets))} detections")

            new_cluster_candidates_and_weights: list[
                tuple[SimpleCluster, Probability]
            ] = []

            logging.info("Num clusters: {len(current_cluster_dist)}")

            new_cluster_candidates_and_weights = computeAllExpansionsAdv(
                current_cluster_dist,
                sources,
                targets,
                assignment_generators,
                num_cores,
                report_progress,
                use_ray,
                truncation_threshold=cutOff,
                max_num_solutions=max_num_solutions,
                mip_method=mip_method,
            )

            # unpack the new cluster candidates with their resampling weights
            new_cluster_candidates, resampling_weights = zip(
                *new_cluster_candidates_and_weights
            )

            assert len(new_cluster_candidates) > 0, "No expansions solutions found!"

            # resample
            sampled_cluster_indices = sample_from_probabilities(
                resampling_weights, num_particles
            )
            unique, count = np.unique(sampled_cluster_indices, return_counts=True)
            histogram = zip(unique, count)

            # choose sampled clusters and correct their size
            new_clusters = []
            for cluster_index, size in histogram:
                cluster = copy.deepcopy(new_cluster_candidates[cluster_index])
                cluster.size = size
                cluster.tracking.id = next_tracking_id
                next_tracking_id += 1
                new_clusters.append(cluster)

            # update cluster distribution
            current_cluster_dist = new_clusters

            # sort clusters w.r.t. size (descending)
            current_cluster_dist = sorted(
                current_cluster_dist, key=lambda c: c.size, reverse=True
            )

            # cut the number of hypothesis
            current_cluster_dist = current_cluster_dist[:max_num_hypotheses]

            for i, c in enumerate(current_cluster_dist):
                c.tracking.id = i

            # all_clusters += [current_cluster_dist]

            for r in reporters:
                r.report_distribution(current_cluster_dist)

            if frame_index % saving_interval == 0 and frame_index > 0:
                for r in reporters:
                    r.close()

    finally:

        for r in reporters:
            r.close()

    return current_cluster_dist


def split_pair_filter(
    source_index, target_index, cdc: ContourDistanceCache, distance_threshold=1.0
):
    # pylint: disable=unused-argument

    return cdc.distance(target_index[:, 0], target_index[:, 1]) < distance_threshold


class SimpleTrackingReporter:
    """Reporting tracking result in simple format"""

    def __init__(
        self, output_folder: str, df: DataFrame, detections, assignment_generators
    ):
        self.all_clusters = []
        self.output_folder = output_folder
        self.df = df
        self.detections = detections
        self.timings = []
        self.timings.append(time.time())
        self.assignment_generators = assignment_generators

        self.final_cluster = None

    def report_distribution(self, current_clusters: list[SimpleCluster]):
        self.final_cluster = current_clusters[0]

    def close(self):
        Path(self.output_folder).mkdir(exist_ok=True)
        edge_list = list(
            map(
                lambda e: (int(e[0]), int(e[1])),
                self.final_cluster.tracking.createIndexTracking().edges,
            )
        )

        tracking_data = [
            dict(
                sourceId=self.detections[edge[0]].id,
                targetId=self.detections[edge[1]].id,
            )
            for edge in edge_list
        ]
        segmentation_data = [
            dict(
                label=cont.label,
                contour=cont.coordinates.tolist(),
                id=cont.id,
                frame=cont.frame,
            )
            for cont in self.detections
        ]

        data_structure = dict(
            segmentation=segmentation_data,
            tracking=tracking_data,
            format_version="0.0.1",
        )

        with open(
            Path(self.output_folder) / "tracking.json", "w", encoding="utf-8"
        ) as output_file:
            json.dump(data_structure, output_file)
        logging.info(self.final_cluster.tracking.createIndexTracking().edges)



""" Configuration for the tracking model """
from __future__ import annotations

import logging
from functools import partial

import numpy as np
import pandas as pd
from scipy.stats import binom, halfnorm, norm
from tensor_walks.impl_np import NP_Impl

from uatrack.assignment import (
    SimpleAssignmentGenerator,
    SimpleContinueGenerator,
    SimpleEndTrackAssGenerator,
    SimpleNewAssGenerator,
    SimpleSplitGenerator,
)
from uatrack.models import (
    ConstantModel,
    ModelExecutor,
    area_growth_computer,
    distance_to_pred_computer,
    first_order_growth,
    predict_property,
    split_dist_computer,
)
from uatrack.utils import ContourDistanceCache, NearestNeighborCache

backend = NP_Impl()


def split_pair_filter(
    source_index, target_index, cdc: ContourDistanceCache, df, distance_threshold=1.0
):
    """Make sure that distances for splitting are not too large"""
    # pylint: disable=unused-argument
    if len(target_index) == 0:
        return np.zeros((len(target_index)), dtype=bool)
    mask = cdc.distance(target_index[:, 0], target_index[:, 1]) < distance_threshold

    return mask


def create_split_children_distance_model(
    data, prob=lambda vs: halfnorm.logsf(vs, loc=0, scale=10)
):
    # pylint: disable=unused-argument
    def split_distance_extractor(
        tracking,
        source_index,
        target_index,
        df,
    ):
        """Extracts the distance between the children of proposed assignments

        Args:
            tracking (_type_): tracking lineage
            source_index (_type_): numpy array of sources for assignments
            target_index (_type_): numpy array of targets for assignments
            df (_type_): dictionary with indexed information

        Returns:
            _type_: euclidean distance between the children of the assignments
        """
        major_extents = df["major_extents"]

        distances = np.zeros(len(target_index), dtype=np.float32)

        for i, (a, b) in enumerate(target_index):
            loc_distances = np.linalg.norm(
                major_extents[a][:, None, :] - major_extents[b][None, :, :], axis=-1
            )

            distances[i] = np.min(loc_distances)

        return distances

    split_children_distance_model = ModelExecutor(
        split_distance_extractor,  # extract the distances
        prob,  # compute the probability
        data,
        "Children split distance",
    )

    return split_children_distance_model


def compute_angle_between_children(maj_extA, maj_extB):
    """Computes the angle between the children

    uses the major axes and the poles in order to orient them accordingly.

    The angle is computed in degrees [0...180]
    """

    loc_distances = np.linalg.norm(maj_extA[:, None, :] - maj_extB[None, :, :], axis=-1)

    min_distance = np.min(loc_distances)
    # print(np.array(np.where(min_distance == loc_distances)).flatten())

    # use the directions pointing towads the new tips
    data = np.array(np.where(min_distance == loc_distances))
    # print(data.shape)
    # print(loc_distances)
    a_ind, b_ind = data[:, 0].flatten()
    adir = maj_extA[1 - a_ind] - maj_extA[a_ind]
    bdir = maj_extB[1 - b_ind] - maj_extB[b_ind]
    # print(adir, bdir)

    return (
        np.arccos(np.dot(adir, bdir) / np.linalg.norm(adir) / np.linalg.norm(bdir))
        * 180
        / np.pi
    )


def log_sum(la, lb):
    """log(a + b) = la + log1p(exp(lb - la)) from https://stackoverflow.com/questions/65233445/how-to-calculate-sums-in-log-space-without-underflow"""
    # la = np.log(a)
    # lb = np.log(b)
    return la + np.log1p(np.exp(lb - la))


def prob_angles(angles, loc, scale):

    # compute the difference
    diff = np.abs(loc - angles)

    # compute the probability of more extreme values
    prob = log_sum(
        norm.logcdf(loc - diff, loc=loc, scale=scale),
        norm.logsf(loc + diff, loc=loc, scale=scale),
    )

    # do not use nan values
    prob[np.isnan(prob)] = -30

    return prob


def prob_cont_angles(angles, scale):

    # compute the probability of more extreme values
    prob = halfnorm.logsf(
        angles, scale=scale
    )  # log_sum(norm.logcdf(loc - diff, loc=loc, scale=scale), norm.logsf(loc + diff, loc=loc, scale=scale))

    # do not use nan values
    prob[np.isnan(prob)] = -30

    return prob


def create_continue_angle_model(data, prob):
    # pylint: disable=unused-argument
    def continue_angle_extractor(
        tracking,
        source_index,
        target_index,
        df,
    ):
        major_axis = df["major_axis"]

        majA = major_axis[source_index.flatten()]
        majB = major_axis[target_index.flatten()]

        vec_norm = np.linalg.norm(majA, axis=-1) / np.linalg.norm(majB, axis=-1)

        raw_values = np.stack(
            [
                [np.dot(a, b) for a, b in zip(majA, majB)] / vec_norm,
                [np.dot(-a, b) for a, b in zip(majA, majB)] / vec_norm,
            ],
            axis=-1,
        )

        raw_values = np.clip(raw_values, -1.0, 1.0)

        angles = np.min(np.arccos(raw_values) * 180 / np.pi, axis=-1)

        return angles

    continue_angle_model = ModelExecutor(continue_angle_extractor, prob, data)

    return continue_angle_model


def create_split_children_angle_model(
    data,
    prob=prob_angles
    #    lambda vs: norm.logpdf(
    #        np.abs(np.cos(vs * (2 * np.pi / 360))),
    #        loc=np.cos(0 * 2 * np.pi / 360),
    #        scale=0.1,
    #    ),
):
    # pylint: disable=unused-argument
    def split_angle_extractor(
        tracking,
        source_index,
        target_index,
        df,
    ):
        major_extents = df["major_extents"]

        cos_angles = np.zeros(len(target_index), dtype=np.float32)

        for i, (a, b) in enumerate(target_index):
            # extract both major axes vectors
            majA = major_extents[a]
            majB = major_extents[b]

            # print(majA)
            # print(majB)

            cos_angles[i] = compute_angle_between_children(majA, majB)

        return cos_angles

    split_children_angle_model = ModelExecutor(split_angle_extractor, prob, data)

    return split_children_angle_model


def create_split_rate_model(data, subsampling):
    def split_rate_comp(values):
        # probs = binom.logpmf(values + 1, n=40, p=10 / 40)
        mean = 70 / subsampling
        scale = 20 / subsampling

        p = 1 - scale / mean
        n = int(mean / p)

        probs = np.log(binom.pmf(values + 1, n=n, p=p))
        indices = np.where(values == -1)
        probs[indices] = np.log(
            0.5
        )  # 50% split probability when we have no parents available
        # print(probs)
        return probs

    # pylint: disable=unused-variable
    split_rate_model = ModelExecutor(
        split_dist_computer,
        # continue_rate_model.model.invert(),
        split_rate_comp,
        data,
    )

    return split_rate_model


def growth_probs(values, mean=1.06, scale=0.3):
    probs = halfnorm.logsf(np.abs(mean - values), scale=scale)
    return probs


def create_growth_model(data, mean, scale):
    return ModelExecutor(
        area_growth_computer,
        # growth_probs,
        partial(growth_probs, mean=mean, scale=scale),
        data,
    )


def create_first_order_growth_model(data, scale, history=1):
    return ModelExecutor(
        partial(first_order_growth, history=history),
        lambda x: log_sum(
            norm.logcdf(x, loc=0, scale=scale), norm.logsf(x, loc=0, scale=scale)
        ),
        data,
        "First order growth",
    )


def distance_to_previous(tracking, source_index, target_index, data):
    # pylint: disable=unused-argument
    all_centroids = data[
        "centroid"
    ]  # np.array(data['centroid'].to_list(), dtype=np.float32)

    sources = all_centroids[source_index]
    targets = all_centroids[target_index]

    distances = np.linalg.norm(sources - targets, axis=-1)

    # if len(distances.shape) == 2:
    # take the sum of distances (only one distance for migration, two distances for division)
    distances = np.sum(distances, axis=-1)

    return distances


def create_continue_keep_position_model(data, prob):
    return ModelExecutor(
        # partial(distance_to_pred_computer, alpha=1, history=0, stop_at_split=True),
        distance_to_previous,
        prob,
        # lambda values: beta.logsf(
        #    np.clip(values.squeeze() / max_distance, 0, 1), a=1, b=3
        # ),
        # lambda values: expon.logsf(values, scale=5),
        data,
    )


def dis_app_prob_func(x_coordinates, max_overshoot, min_dist, x_size: int):
    total_auc = 0.5 * (min_dist + max_overshoot) ** 2
    area_func = lambda x: 0.5 * (x + max_overshoot) ** 2
    prob = np.log(
        1
        - area_func(
            np.clip(
                np.min(np.stack([x_coordinates, x_size - x_coordinates]), axis=0),
                -max_overshoot,
                min_dist,
            )
        )
        / total_auc
    )

    prob[np.isnan(prob)] = -30

    return prob


# pylint: disable=unused-argument
def source_position_extractor(
    tracking,
    source_index,
    target_index,
    df,
):
    # extract the x-position of all sources
    return df["centroid"][source_index.flatten()][:, 1]


def create_disappear_model(data, width: int):
    return ModelExecutor(
        source_position_extractor,
        partial(dis_app_prob_func, max_overshoot=20, min_dist=20, x_size=width),
        data,
    )


def create_continue_temp_position_model(data, prob, history=3):
    return ModelExecutor(
        partial(
            distance_to_pred_computer, alpha=None, history=history, stop_at_split=True
        ),
        prob,
        data,
        "First order position model",
    )


def create_split_movement_model(data, prob):
    return ModelExecutor(
        distance_to_previous,
        prob,
        data,
    )


def migration_prob(data: dict[str, np.ndarray], subsampling: int):
    """Model for cell migration probability (tailored to specific sequence).

    Args:
        data (dict[str, np.ndarray]): single-cell data
        subsampling (int): subsampling factor

    Returns:
        _type_: _description_
    """

    # 1. data < 0: no division has been observed yet -> cell could divide at any moment (p=0.5)
    # 2. 0 <= data < 40: cell is quite young (p=0.9)
    # 3. 40 <= data < 81: cell is in division age (p=0.1)
    # 4. 81 <= data: cell is over typical division age. We do not know what's going on (p=0.5)
    print(f"data dim: {data.shape}")
    # expected division timeframe between [min_time, max_time]. Divided by subsampling factor
    min_time = 40 / subsampling
    max_time = 80 / subsampling

    # probability of migration is high before division time window. During division time window it is low. After that it is unknown (0.5)
    return (
        (data < 0) * 0.5
        + ((0 <= data) & (data < min_time)) * 0.9
        + ((min_time <= data) & (data <= max_time)) * 0.1
        + (max_time < data) * 0.5
    )


def create_age_migration_model(
    data: dict[str, np.ndarray], subsampling: int
) -> ModelExecutor:
    """Creates migration probability model based on cell age

    Args:
        data (dict[str, np.ndarray]): single-cell data
        subsampling (int): subsampling factor

    Returns:
        ModelExecutor: scoring model
    """
    return ModelExecutor(
        split_dist_computer, lambda x: np.log(migration_prob(x, subsampling)), data
    )


def create_age_division_model(
    data: dict[str, np.ndarray], subsampling: int
) -> ModelExecutor:
    """Create age division model.
       This probability is derived from the probability not to migrate

    Args:
        data (dict[str, np.ndarray]): single-cell data
        subsampling (int): subsamping factor

    Returns:
        ModelExecutor: scoring model executor
    """
    return ModelExecutor(
        split_dist_computer, lambda x: np.log(1 - migration_prob(x, subsampling)), data
    )


def create_end_pos_model(
    data: dict[str, np.ndarray], width: int, subsampling: int, history=1
) -> ModelExecutor:
    """Create model for predicting cell leaving the field of view. This lowers cell disappearance probabilites at the center of the microscopy image (and far from the vertical edges)

    Args:
        data (dict[str, np.ndarray]): single-cell data
        width (int): width of the image
        subsampling (int): subsampling factor
        history (int, optional): number of historic cell movements to use for position prediction. Defaults to 1.
    """

    def position(tracking, source_index, target_index, df):
        # pylint: disable=unused-argument

        # predict cell movement
        source_predictions = predict_property(
            tracking,
            source_index=source_index,
            df=df,
            history=history,
            stop_at_split=True,
            property_name="centroid",
        )

        return source_predictions

    def pos_log_pdf(values):
        values = values[:, 0]
        min_dist = np.minimum(values, np.abs(width - values))

        # all cells have a standard disappear probability
        p = np.zeros_like(values) + np.log(0.25)
        # all cells far from the border have a lower disappear prob
        p[min_dist > np.clip(10 * subsampling, 0, 100)] = -4  # -6 #np.log(1e-6)

        return p

    return ModelExecutor(position, pos_log_pdf, data)


class SimpleCDC:
    """class to compute contour distances (approximated by center position in favor of speed)"""

    def __init__(self, positions):
        self.positions = positions

    def distance(self, indA, indB):
        return np.linalg.norm(self.positions[indA] - self.positions[indB], axis=1)


def create_constant_disappear_model(app_prob: float):
    return ConstantModel(np.log(app_prob)), ConstantModel(np.log(app_prob))


def standard_models(
    data,
    subsampling,
    use_long_hist=0,
    app_prob=0.25,
    mig_growth_scale=0.05,
    mig_movement_scale=20,
    div_growth_scale=None,
    div_movement_scale=20,
    mig_growth_mean=1.008,
    div_growth_mean=1.016,
):

    if div_growth_scale is None:
        div_growth_scale = 2 * mig_growth_scale
    if div_movement_scale is None:
        div_movement_scale = 2 * mig_movement_scale

    # dis (-appear) models
    constant_new_model, constant_end_model = create_constant_disappear_model(app_prob)

    # Movement models
    continue_movement_model = create_continue_temp_position_model(
        data,
        prob=lambda val: halfnorm.logsf(
            val.flatten(), scale=mig_movement_scale * subsampling
        ),
        history=use_long_hist,
    )
    # continue_movement_model = create_continue_keep_position_model(
    #    data, prob=lambda val: halfnorm.logsf(val, scale=mig_movement_scale*subsampling)
    # )
    split_movement_model = create_split_movement_model(
        data,
        prob=lambda val: halfnorm.logsf(val, scale=div_movement_scale * subsampling),
    )

    # growth models:
    continue_growth_model = create_growth_model(
        data, mean=mig_growth_mean**subsampling, scale=mig_growth_scale * subsampling
    )
    split_growth_model = create_growth_model(
        data, mean=div_growth_mean**subsampling, scale=div_growth_scale * subsampling
    )

    return (
        [constant_new_model],
        [constant_end_model],
        [continue_movement_model, continue_growth_model],
        [split_movement_model, split_growth_model],
    )


def create_filters(
    data: dict[str, np.ndarray], df: pd.DataFrame, k=15
) -> tuple[list, SimpleCDC]:
    """Creates a nearest neighbor filter and a distance computer

    Args:
        data (dict[str, np.ndarray]): single-cell data
        df (pd.DataFrame): single-cell data (in table format)
        k (int, optional): Number of nearest neighbors to consider. Defaults to 15.

    Returns:
        tuple[list, SimpleCDC]: tuple of nearest neighbor filter and distance computer
    """

    positions = data["centroid"]
    cdc = SimpleCDC(positions)

    logging.info("Compute nearest neighbor cache")
    nnc = NearestNeighborCache(df)

    filters = [lambda s, t: nnc.kNearestNeigborsMatrixMask(k, s, t)]

    return filters, cdc


def use_disappear_model(
    data: dict[str, np.ndarray], subsampling: int, width: int, **kwargs
) -> tuple[
    list[ModelExecutor], list[ModelExecutor], list[ModelExecutor], list[ModelExecutor]
]:
    """Adds disappear models to the first-order models

    Args:
        data (dict[str, np.ndarray]): single-cell data (numpy arrays)
        subsampling (int): subsampling factor. Used to linearly scale the scale value of statistical distributions
        width (int): width of the microscopy image (in pixels)

    Returns:
        tuple[list[ModelExecutor], list[ModelExecutor], list[ModelExecutor], list[ModelExecutor]]: list of scoring models for appearance, disappearance, migration and division models
    """

    # make standard models
    (
        constant_new_models,
        _,
        migration_models,
        split_models,
    ) = use_first_order_model(data=data, subsampling=subsampling, **kwargs)

    end_pos_model = create_end_pos_model(data, width, subsampling)

    return constant_new_models, [end_pos_model], migration_models, split_models


def add_split_distance_model(
    data: dict[str, np.ndarray], subsampling: int, **kwargs
) -> tuple[
    list[ModelExecutor], list[ModelExecutor], list[ModelExecutor], list[ModelExecutor]
]:
    """Adds division distance model to the first-order models

    Args:
        data (dict[str, np.ndarray]): single-cell data (numpy arrays)
        subsampling (int): subsampling factor. Used to linearly scale the scale value of statistical distributions

    Returns:
        tuple[list[ModelExecutor], list[ModelExecutor], list[ModelExecutor], list[ModelExecutor]]: list of scoring models for appearance, disappearance, migration and division models
    """

    # make first-order models
    (
        constant_new_models,
        constant_end_models,
        migration_models,
        split_models,
    ) = use_first_order_model(data=data, subsampling=subsampling, **kwargs)

    # add children distance model
    split_children_distance_model = create_split_children_distance_model(
        data, prob=lambda vs: halfnorm.logsf(vs, loc=0, scale=3 * subsampling)
    )
    split_models += [split_children_distance_model]

    return constant_new_models, constant_end_models, migration_models, split_models


def add_angle_models(
    data: dict[str, np.ndarray], subsampling: int, **kwargs
) -> tuple[
    list[ModelExecutor], list[ModelExecutor], list[ModelExecutor], list[ModelExecutor]
]:
    """Adds migration and division orientation models to the first-order models

    Args:
        data (dict[str, np.ndarray]): single-cell data (numpy arrays)
        subsampling (int): subsampling factor. Used to linearly scale the scale value of statistical distributions

    Returns:
        tuple[list[ModelExecutor], list[ModelExecutor], list[ModelExecutor], list[ModelExecutor]]: list of scoring models for appearance, disappearance, migration and division models
    """

    # make first-order models
    (
        constant_new_models,
        constant_end_models,
        migration_models,
        split_models,
    ) = use_first_order_model(data=data, subsampling=subsampling, **kwargs)

    continue_angle_model = create_continue_angle_model(
        data, prob=partial(prob_cont_angles, scale=20 * subsampling)
    )
    split_children_angle_model = create_split_children_angle_model(
        data, prob=partial(prob_angles, loc=135, scale=20 * subsampling)
    )

    # add models
    migration_models += [continue_angle_model]
    split_models += [split_children_angle_model]

    return constant_new_models, constant_end_models, migration_models, split_models
    # return make_assignmet_generators(df, data, constant_new_model, constant_end_model, migration_models, split_models)


def add_age_models(
    data: dict[str, np.ndarray], subsampling: int, **kwargs
) -> tuple[
    list[ModelExecutor], list[ModelExecutor], list[ModelExecutor], list[ModelExecutor]
]:
    """Adds the age model to the first-order models

    Args:
        data (dict[str, np.ndarray]): single-cell data (numpy arrays)
        subsampling (int): subsampling factor. Used to linearly scale the scale value of statistical distributions

    Returns:
        tuple[list[ModelExecutor], list[ModelExecutor], list[ModelExecutor], list[ModelExecutor]]: list of scoring models for appearance, disappearance, migration and division models
    """

    # make first-order models
    (
        constant_new_models,
        constant_end_models,
        migration_models,
        split_models,
    ) = use_first_order_model(data=data, subsampling=subsampling, **kwargs)

    # create age models
    migration_age_model, split_age_model = (
        create_age_migration_model(data, subsampling),
        create_age_division_model(data, subsampling),
    )

    # add models
    migration_models += [migration_age_model]
    split_models += [split_age_model]

    return constant_new_models, constant_end_models, migration_models, split_models


def use_nearest_neighbor(
    data: dict[str, np.ndarray],
    subsampling: int,
    mig_growth_scale=0.05,
    mig_movement_scale=20,
    div_growth_scale: float = None,
    div_movement_scale=20.0,
    **kwargs,
) -> tuple[
    list[ModelExecutor], list[ModelExecutor], list[ModelExecutor], list[ModelExecutor]
]:
    """Nearest neighbor tracking configuration assumes that cells do not move or grow and builds statistical distributions to score cell development based on this assumption.

    Args:
        data (dict[str, np.ndarray]): single-cell data (numpy arrays)
        subsampling (int): subsampling factor. Used to linearly scale the scale value of statistical distributions
        mig_growth_scale (float, optional): scale parameter for migration growth. Defaults to 0.05.
        mig_movement_scale (_type_, optional): scale parameter for migration movement (in pixels). Defaults to 20..
        div_growth_scale (float, optional): scale parameter for division growth. Defaults to None.
        div_movement_scale (_type_, optional): scale parameter for division movement. Defaults to 20.

    Returns:
        tuple[list[ModelExecutor], list[ModelExecutor], list[ModelExecutor], list[ModelExecutor]]: list of scoring models for appearance, disappearance, migration and division models
    """

    if div_growth_scale is None:
        div_growth_scale = 2 * mig_growth_scale
    if div_movement_scale is None:
        div_movement_scale = 2 * mig_movement_scale

    # make standard models
    new_models, end_models, _, _ = standard_models(
        data=data, subsampling=subsampling, **kwargs
    )

    # Movement models: with history=0 --> Makes them expect no movement -> look for cell association at the same position
    continue_movement_model = create_continue_temp_position_model(
        data,
        prob=lambda val: halfnorm.logsf(
            val.flatten(), scale=mig_movement_scale * subsampling
        ),
        history=0,
    )
    split_movement_model = create_continue_temp_position_model(
        data,
        prob=lambda val: halfnorm.logsf(val, scale=div_movement_scale * subsampling),
        history=0,
    )

    # growth models: Assume no growth (simplification) --> look for cells with same size
    continue_growth_model = create_growth_model(
        data, mean=1, scale=mig_growth_scale * subsampling
    )
    split_growth_model = create_growth_model(
        data, mean=1, scale=div_growth_scale * subsampling
    )

    migration_models = [continue_movement_model, continue_growth_model]
    split_models = [split_movement_model, split_growth_model]

    return new_models, end_models, migration_models, split_models


def use_first_order_model(
    data: dict[str, np.ndarray],
    subsampling: int,
    use_long_hist=1,
    fo_mov_scale=5,
    fo_growth_scale=10,
    **kwargs,
) -> tuple[
    list[ModelExecutor], list[ModelExecutor], list[ModelExecutor], list[ModelExecutor]
]:
    """Create first order-tracking configuration models.
       Idea: movement and growth are predicted based on past cell movement and growth.

    Args:
        data (dict[str, np.ndarray]): single-cell data (numpy arrays)
        subsampling (int): subsampling factor. Used to linearly scale the scale value of statistical distributions
        use_long_hist (int, optional): the track length used to predict single-cell properties. Defaults to 1.
        fo_mov_scale (float, optional): scale parameter for half-normal movement distribution. Defaults to 5.
        fo_growth_scale (float, optional): scale parameter for half-normal growth distribution. Defaults to 10.

    Returns:
        tuple[list[ModelExecutor], list[ModelExecutor], list[ModelExecutor], list[ModelExecutor]]: list of scoring models for appearance, disappearance, migration and division models
    """

    # make standard models
    constant_new_models, constant_end_models, _, _ = standard_models(
        data=data, subsampling=subsampling, **kwargs
    )

    # First order movement models
    continue_movement_model = create_continue_temp_position_model(
        data,
        lambda val: halfnorm.logsf(val, scale=20 + fo_mov_scale * subsampling),
        history=use_long_hist,
    )  # first order movement model
    continue_growth_model = create_first_order_growth_model(
        data, 50 + fo_growth_scale * subsampling, history=use_long_hist
    )  # None # first order growth model

    # Division models (only look at children distance & growth) -> divisions are restricted by combinations
    split_movement_model = create_continue_temp_position_model(
        data,
        lambda val: halfnorm.logsf(
            val, loc=0, scale=20 + 2 * fo_mov_scale * subsampling
        ),
        history=use_long_hist,
    )  # create_continue_temp_position_model(data, lambda val: halfnorm.logsf(val, scale=10*subsampling), history=1) #create_split_children_distance_model(data, prob=lambda vs: halfnorm.logsf(vs, loc=0, scale=3*subsampling)) #None # first order movement model
    split_growth_model = create_first_order_growth_model(
        data, 50 + 2 * fo_growth_scale * subsampling, history=use_long_hist
    )  # first order growth model (e.g. $(sum child area)=$(historical growth rate) * 2)

    # collect models
    migration_models = [
        continue_movement_model,
        continue_growth_model,
    ]
    split_models = [
        split_movement_model,
        split_growth_model,
    ]

    return constant_new_models, constant_end_models, migration_models, split_models


def add_growth_model(
    data,
    subsampling,
    mig_growth_scale=0.05,
    div_growth_scale=None,
    mig_growth_mean=1.008,
    div_growth_mean=1.016,
    **kwargs,
):

    new_models, end_models, migration_models, split_models = use_first_order_model(
        data=data, subsampling=subsampling, **kwargs
    )

    if div_growth_scale is None:
        div_growth_scale = 2 * mig_growth_scale

    # growth models:
    continue_growth_model = create_growth_model(
        data, mean=mig_growth_mean**subsampling, scale=mig_growth_scale * subsampling
    )
    split_growth_model = create_growth_model(
        data, mean=div_growth_mean**subsampling, scale=div_growth_scale * subsampling
    )

    # replace first-order growth models by bioligacally motivated models
    migration_models[1] = continue_growth_model
    split_models[1] = split_growth_model

    return new_models, end_models, migration_models, split_models


def make_assignmet_generators(
    df: pd.DataFrame,
    data: dict[str, np.ndarray],
    constant_new_models: list[ModelExecutor],
    constant_end_models: list[ModelExecutor],
    migration_models: list[ModelExecutor],
    split_models: list[ModelExecutor],
    max_distance=200,
) -> list[SimpleAssignmentGenerator]:
    """Creates a list of assignment generators based on the scoring models and basic nearest neighbor filtering

    Args:
        df (pd.DataFrame): single-cell data (table)
        data (dict[str, np.ndarray]): single-cell data (numpy arrays)
        constant_new_models (list[ModelExecutor]): statistical models for cell appearance
        constant_end_models (list[ModelExecutor]): statistical models for cell disappearance
        migration_models (list[ModelExecutor]): statistical models for cell migration
        split_models (list[ModelExecutor]): statistical models for cell division
        max_distance (int, optional): maximum distance of cells (in pixel) to be considered in assignments. Defaults to 200.

    Returns:
        list[SimpleAssignmentGenerator]: assignment generators
    """
    filters, cdc = create_filters(data, df)

    # create assignment generators
    return [
        SimpleNewAssGenerator(constant_new_models),
        SimpleEndTrackAssGenerator(constant_end_models),
        SimpleContinueGenerator(filters, migration_models),
        SimpleSplitGenerator(
            filters,
            partial(split_pair_filter, cdc=cdc, distance_threshold=max_distance, df=df),
            split_models,
        ),
    ]


def setup_assignment_generators(
    df: pd.DataFrame, subsampling_factor: int, model: str = "NN"
) -> list[SimpleAssignmentGenerator]:
    """Create the assignment generators and cell models for scoring tracking assignments

    Args:
        df (pd.DataFrame): data frame with single-cell data from segementation
        subsampling_factor (int): subsampling factor
        model (str, optional): the model configuration for the tracking. Defaults to "NN".

    Returns:
        list[SimpleAssignmentGenerator]: List of assignment generators equipped with cell models
    """

    # arrange single-cell information into numpy arrays (greatly increases the speed, as data can be immediately indexed)
    data = {
        "area": np.array(df["area"].to_list(), dtype=np.float32),
        "centroid": np.array(df["centroid"].to_list(), dtype=np.float32),
        "major_extents": np.array(df["major_extents"].to_list(), dtype=np.float32),
        "major_axis": np.array(df["major_axis"].to_list(), dtype=np.float32),
    }

    subsampling = subsampling_factor

    # create biologically motivated models
    if model == "NN":
        (
            constant_new_models,
            constant_end_models,
            migration_models,
            split_models,
        ) = use_nearest_neighbor(data=data, subsampling=subsampling_factor)
    elif model == "FO":
        (
            constant_new_models,
            constant_end_models,
            migration_models,
            split_models,
        ) = use_first_order_model(data=data, subsampling=subsampling_factor)
    elif model == "FO+G":
        (
            constant_new_models,
            constant_end_models,
            migration_models,
            split_models,
        ) = add_growth_model(
            data=data, subsampling=subsampling_factor
        )  # add_angle_models(data=data, subsampling=subsampling_factor) #use_first_order_model(data=data, subsampling=subsampling_factor) #use_nearest_neighbor(data=data, subsampling=subsampling_factor) #
    elif model == "FO+O":
        (
            constant_new_models,
            constant_end_models,
            migration_models,
            split_models,
        ) = add_angle_models(data=data, subsampling=subsampling_factor)
    elif model == "FO+DD":
        # make standard models
        (
            constant_new_models,
            constant_end_models,
            migration_models,
            split_models,
        ) = use_first_order_model(data=data, subsampling=subsampling)

        # add children distance model
        split_children_distance_model = create_split_children_distance_model(
            data, prob=lambda vs: halfnorm.logsf(vs, loc=0, scale=3 * subsampling)
        )
        split_models += [split_children_distance_model]

    elif model == "FO+G+O+DD":

        (
            constant_new_models,
            constant_end_models,
            migration_models,
            split_models,
        ) = add_growth_model(data=data, subsampling=subsampling_factor)

        # add children distance model
        split_children_distance_model = create_split_children_distance_model(
            data, prob=lambda vs: halfnorm.logsf(vs, loc=0, scale=3 * subsampling)
        )
        split_models += [split_children_distance_model]

        # add orientation models
        continue_angle_model = create_continue_angle_model(
            data, prob=partial(prob_cont_angles, scale=20 * subsampling)
        )
        split_children_angle_model = create_split_children_angle_model(
            data, prob=partial(prob_angles, loc=135, scale=20 * subsampling)
        )

        migration_models += [continue_angle_model]
        split_models += [split_children_angle_model]

    # create the assignment candidate generators
    assignment_generators = make_assignmet_generators(
        df=df,
        data=data,
        constant_new_models=constant_new_models,
        constant_end_models=constant_end_models,
        migration_models=migration_models,
        split_models=split_models,
    )

    return assignment_generators



""" Collection of utility functionality """

from __future__ import annotations

import gzip
import itertools
import json
import multiprocessing
from functools import partial, reduce
from itertools import tee
from pathlib import Path

import cv2
import networkx as nx
import numpy as np
import ray
from acia.base import Overlay
from acia.segm.formats import parse_simple_segmentation
from acia.tracking.formats import parse_simple_tracking
from acia.tracking.utils import subsample_lineage
from acia.viz import VideoExporter2
from pandas import DataFrame
from PIL import Image, ImageDraw
from scipy.spatial.distance import cdist
from tqdm.auto import tqdm
from tqdm.contrib.concurrent import process_map


def product(list1):
    return reduce((lambda x, y: x * y), list1)


def distance(point_a, point_b):
    """
    point_a: (x,y, ...)
    point_b: (x,y, ...)
    """

    assert len(point_a) == len(point_b)

    return np.linalg.norm(np.array(point_a) - np.array(point_b), ord=2)


class DistanceComputer:
    """Compute and cache distances"""

    def __init__(self):
        """
        Precompute all possible distances
        """
        self._distance_lookup = {}

    def precompute(self, detections, num_cores=1):
        if num_cores == 1:
            self.precomputeDistances(detections)
        else:
            self.precomputeDistancesParallel(detections, num_cores)

    def precomputeDistances(self, detections):
        self._distance_lookup = {}
        detections = list(detections)
        max_frame = max(map(lambda det: det.frame_id, detections))
        results = itertools.chain.from_iterable(
            map(
                lambda frame_id: DistanceComputer.precomputeDistancesInFrame(
                    detections, frame_id
                ),
                range(0, max_frame + 1),
            )
        )
        for key, value in results:
            self._distance_lookup[key] = value

    @staticmethod
    def precomputeDistancesInFrame(detections, frame_id):
        result = []
        frame_detections = filter(lambda det: det.frame_id == frame_id, detections)
        for det1, det2 in itertools.combinations(frame_detections, 2):
            result.append(
                (tuple(sorted([det1.id, det2.id])), det1.polygon.distance(det2.polygon))
            )

        return result

    @staticmethod
    def distance_comp(det_set):
        assert len(det_set) == 2
        det1, det2 = det_set[0], det_set[1]
        return (tuple(sorted([det1.id, det2.id])), det1.polygon.distance(det2.polygon))

    def precomputeDistancesParallel(self, detections, num_cores):
        self._distance_lookup = {}
        detections = list(detections)
        max_frame = max(map(lambda det: det.frame_id, detections))
        with multiprocessing.Pool(num_cores) as p:

            result = []
            for ind_result in tqdm(
                p.imap_unordered(
                    lambda frame_id: DistanceComputer.precomputeDistancesInFrame(
                        detections, frame_id
                    ),
                    reversed(range(0, max_frame + 1)),
                )
            ):
                result.append(ind_result)

            p.close()
            p.join()
            p.terminate()

        # result = p.map(DistanceComputer.distance_comp, itertools.combinations(detections, 2))

        for key, value in itertools.chain.from_iterable(result):
            self._distance_lookup[key] = value

    def distance(self, det1, det2):
        key = tuple(sorted([det1.id, det2.id]))
        if key not in self._distance_lookup:
            self._distance_lookup[key] = det1.polygon.distance(det2.polygon)
        return self._distance_lookup[key]


def pair_ind_to_dist_ind(d, i, j):
    assert np.all(i < j)
    index = d * (d - 1) / 2 - (d - i) * (d - i - 1) / 2 + j - i - 1
    return index.astype(np.uint32)


def compute_distances(i, contours):
    local_distances = []

    for j in range(i + 1, len(contours)):
        local_distances.append(np.min(cdist(contours[i], contours[j])))

    return local_distances


@ray.remote
def ray_compute_distances(i, contours):
    return compute_distances(i, contours)


class ContourDistanceCache:
    """Cache for quickly computing the contour distance cache"""

    def __init__(self, df):
        self.n = len(df)

        contours = df["contour"].to_numpy()

        distances = []
        distances = self.__multi_processing(contours)
        # with multiprocessing.Pool(16) as pool:
        #    distances = list(tqdm(pool.imap(partial(compute_distances(contours=contours)),range(self.n)), total=self.n))
        #    #for i in tqdm(range(self.n)):

        self.distances = np.concatenate(distances)

        print(self.distances.shape)

    def __multi_processing(self, contours):
        return process_map(
            partial(compute_distances, contours=contours), range(self.n), chunksize=10
        )

    def __ray_processing(self, contours):
        # pylint: disable=unused-private-member
        distances = [ray_compute_distances.remote(i, contours) for i in range(self.n)]

        def to_iterator(obj_ids):
            while obj_ids:
                done, obj_ids = ray.wait(obj_ids)
                yield ray.get(done[0])

        ray_it = to_iterator(distances)

        # if report_progress:
        ray_it = tqdm(ray_it, total=len(distances))

        return list(ray_it)

        # return ray_compute_distances()

    def distance(self, source_index, target_index):
        return self.distances[pair_ind_to_dist_ind(self.n, source_index, target_index)]


class NearestNeighborCache:
    """Cache for fast computation of nearest neighbor sets"""

    def __init__(self, df):
        self.df = df

        self.positions = np.array(df["centroid"].tolist(), dtype=np.float32)

    @property
    def num_elements(self):
        return len(self.positions)

    def computePossibleTargets(self, targetFrame: int):
        # first filter data frame for frame targets
        return self.df[self.df["frame"] == targetFrame].index.to_numpy()

    def neighborDistances(self, listOfSources, possible_targets):

        # repeat the targets for the number of sources
        stacked_targets = np.tile(possible_targets, (len(listOfSources),))
        # repeat the sources for the number of targets
        stacked_sources = np.repeat(listOfSources, len(possible_targets))

        # print(stacked_sources, stacked_targets)

        # compute the right indices
        # indices = pair_ind_to_dist_ind(self.num_elements, stacked_sources, stacked_targets)

        # lookup the distances
        # return self.distanceCache.pairwise_distance[indices].reshape(-1, len(possible_targets))
        return self.distance(stacked_sources, stacked_targets).reshape(
            -1, len(possible_targets)
        )

    def neighborDistancesMatrix(self, sourceIndices, targetIndexMatrix):
        num_targets = targetIndexMatrix.shape[1]

        # repeat the sources for the number of targets
        stacked_sources = np.repeat(sourceIndices, num_targets)
        flattenedTargets = targetIndexMatrix.flatten()

        # compute the right indices
        # indices = pair_ind_to_dist_ind(self.num_elements, stacked_sources, flattenedTargets)

        # lookup the distances
        return self.distance(stacked_sources, flattenedTargets).reshape(-1, num_targets)

    def sortedNeighbors(self, listOfSources, targetFrame):
        """
        listOfSources: n source indices
        targetFrame: int

        returns nxk index map where k is the number of detections in the target frame
        """
        # identify all detections in target frame
        possible_targets = self.computePossibleTargets(targetFrame)

        distances = self.neighborDistances(listOfSources, possible_targets)

        # sort them by indices
        return possible_targets[np.argsort(distances).flatten()].reshape(
            len(listOfSources, len(possible_targets))
        )

    def kNearestNeighborsFrame(self, k: int, listOfSources, targetFrame):
        # identify all detections in target frame
        possible_targets = self.computePossibleTargets(targetFrame)

        if len(possible_targets) <= k:
            # if we have lesseq than k detections finding k-nearest neighbors is trivial
            return np.tile(possible_targets, (len(listOfSources),)).reshape(
                (-1, len(possible_targets))
            )

        # compute distances
        distances = self.neighborDistances(listOfSources, possible_targets)

        # print(distances)

        # sort them by indices
        return possible_targets[np.argpartition(distances, k)[:, :k].flatten()].reshape(
            (len(listOfSources), k)
        )

    def kNearestNeigbors(self, k: int, sourceIndices, targetIndices):
        distances = self.neighborDistancesMatrix(sourceIndices, targetIndices)

        # print('dist shape', distances.shape)

        return np.argpartition(distances, k)[:, :k]  # .reshape((len(listOfSources), k))

    def kNearestNeigborsMatrixMask(self, k: int, sourceIndices, targetIndexMatrix):
        if targetIndexMatrix.shape[1] <= k:
            # if we have lesseq than k detections finding k-nearest neighbors is trivial
            mask = np.ones_like(targetIndexMatrix, dtype=bool)

        else:
            correctIndices = self.kNearestNeigbors(k, sourceIndices, targetIndexMatrix)

            # print(correctIndices.shape)

            # convert the filtered index list into a mask on the full target index matrix
            mask = np.zeros_like(targetIndexMatrix)
            # writing the true values is bit complex due to selection
            mask[np.arange(sourceIndices.shape[0])[:, None], correctIndices] = True

        return mask

    def distance(self, indexListA, indexListB):
        distances = np.linalg.norm(
            self.positions[indexListA] - self.positions[indexListB], axis=-1
        )
        return distances
        # return self.distanceCache.pairwise_distance[pair_ind_to_dist_ind(self.num_elements, indexListA, indexListB)]


def load_local_data(simple_seg_file):
    print("Loading segmentation...")
    # Load segmentation or additional tracking information
    with open(simple_seg_file, encoding="UTF-8") as input_file:
        ov = parse_simple_segmentation(input_file.read())

    return ov


def load_single_cell_information(input_file: Path) -> DataFrame:

    # all_detections, width, ov = load_local_data(output_folder/ "pred_simpleSegmentation.json.gz", width=997, subsampling_factor=40) #load_omero_data(output_folder, omero_id)
    ov = load_local_data(input_file)
    # ov = Overlay([cont for cont in ov if cont.frame <= 100])
    all_detections = ov.contours

    # split_proposals, major_axes = compute_split_proposals(all_detections)

    # create the main data frame
    entries = []
    for _, det in enumerate(all_detections):
        # compute the axis info
        (
            width,
            length,
            major_axis,
            minor_axis,
            major_extents,
            minor_extents,
        ) = compute_axes_info(det.polygon)

        # add entry for this cell
        entries.append(
            {
                "area": det.area,
                "centroid": np.array(det.center),
                "perimeter": det.polygon.length,
                "id": det.id,
                "frame": det.frame,
                "contour": np.array(det.coordinates),
                "width": width,
                "length": length,
                "major_axis": major_axis,
                "minor_axis": minor_axis,
                "major_extents": major_extents,
                "minor_extents": minor_extents,
            }
        )
    df = DataFrame(entries)

    return df, all_detections


def extract_single_cell_information(ov: Overlay) -> DataFrame:

    # ov = Overlay([cont for cont in ov if cont.frame <= 100])
    all_detections = ov.contours

    # split_proposals, major_axes = compute_split_proposals(all_detections)

    # create the main data frame
    entries = []
    for _, det in enumerate(all_detections):
        # compute the axis info
        (
            width,
            length,
            major_axis,
            minor_axis,
            major_extents,
            minor_extents,
        ) = compute_axes_info(det.polygon)

        # add entry for this cell
        entries.append(
            {
                "area": det.area,
                "centroid": np.array(det.center),
                "perimeter": det.polygon.length,
                "id": det.id,
                "frame": det.frame,
                "contour": np.array(det.coordinates),
                "width": width,
                "length": length,
                "major_axis": major_axis,
                "minor_axis": minor_axis,
                "major_extents": major_extents,
                "minor_extents": minor_extents,
            }
        )
    df = DataFrame(entries)

    return df, all_detections


def compute_axes_info(polygon):
    """Extract information about the major and minor axes from a single detection

    Args:
        det (_type_): detection object

    Returns:
        _type_: a lot of information about major and minor axes
    """
    # get the minimum rotated rectangle around a detection
    mrr = polygon.minimum_rotated_rectangle
    coordinates = np.array(mrr.boundary.coords)

    # substract all coordinates to get the distances
    diff_vectors = coordinates[:-1] - coordinates[1:]
    distances = np.linalg.norm(diff_vectors, axis=1)

    # minor axis has lower distance than major axis
    minor_index = np.argmin(distances)
    major_index = np.argmax(distances)

    # compute width and length (minor axis, major axis)
    width = distances[minor_index]
    length = distances[major_index]

    # extract minor and major axis
    minor_axis = diff_vectors[minor_index]
    major_axis = diff_vectors[major_index]

    center = np.array(mrr.centroid.coords)[0]

    # construct axis endpoints (extents)
    major_extents = center + 0.5 * np.array([major_axis, -major_axis])
    minor_extents = center + 0.5 * np.array([minor_axis, -minor_axis])

    # return all extracted values
    return width, length, major_axis, minor_axis, major_extents, minor_extents


def save_tracking(final_cluster, detections, output_file: Path = "tracking.json.gz"):
    # Path(self.output_folder).mkdir(exist_ok=True)
    edge_list = list(
        map(
            lambda e: (int(e[0]), int(e[1])),
            final_cluster.tracking.createIndexTracking().edges,
        )
    )

    tracking_data = [
        dict(
            sourceId=detections[edge[0]].id,
            targetId=detections[edge[1]].id,
        )
        for edge in edge_list
    ]
    segmentation_data = [
        dict(
            label=cont.label,
            contour=cont.coordinates.tolist(),
            id=cont.id,
            frame=cont.frame,
        )
        for cont in detections
    ]

    data_structure = dict(
        segmentation=segmentation_data,
        tracking=tracking_data,
        format_version="0.0.1",
    )

    with gzip.open(output_file, "wt") as output_writer:
        json.dump(data_structure, output_writer)


def render_tracking_video(
    image_source,
    overlay_iterator,
    tracking,
    output_file="track.mp4",
    framerate=3,
):
    """Render tracking video

    Args:
        image_source (_type_): acia image source iterator
        overlay_iterator (_type_): iterator over cell overlay
        tracking (_type_): tracking graph
        output_file (str, optional): Output file for the video. Defaults to "track.avi".
        framerate (int, optional): Rendering framerate (fps). Defaults to 3.
    """

    overlay_iterator, consumer = tee(overlay_iterator)
    all_contours = reduce(
        lambda a, b: a + b,
        [[cont for overlay in iter(consumer) for cont in overlay]],
        [],
    )
    contour_lookup = {cont.id: cont for cont in all_contours}

    with VideoExporter2.fast_vp9(str(output_file), framerate) as ve:
        for frame, (image, overlay) in enumerate(
            tqdm(zip(image_source, overlay_iterator))
        ):
            pil_image = Image.fromarray(image.raw)
            overlay.draw(pil_image, "#00FFFF", None)

            draw = ImageDraw.Draw(pil_image)

            draw.text((30, 30), f"Frame: {frame}", fill="white")

            np_image = np.array(pil_image)

            for cont in overlay:
                if cont.id in tracking.nodes:
                    edges = tracking.out_edges(cont.id)

                    born = tracking.in_degree(cont.id) == 0

                    for edge in edges:
                        source = contour_lookup[edge[0]].center
                        target = contour_lookup[edge[1]].center

                        line_color = (0, 0, 255)  # bgr: red

                        if len(edges) > 1:
                            line_color = (255, 0, 0)  # bgr: blue

                        cv2.line(
                            np_image,
                            tuple(map(int, source)),
                            tuple(map(int, target)),
                            line_color,
                            thickness=3,
                        )

                        if born:
                            cv2.circle(
                                np_image,
                                tuple(map(int, source)),
                                3,
                                (203, 192, 255),
                                thickness=1,
                            )

                    if len(edges) == 0:
                        cv2.rectangle(
                            np_image,
                            cont.center.astype(np.int32) - 2,
                            cont.center.astype(np.int32) + 2,
                            (203, 192, 255),
                        )

                        # draw.line([tuple(source), tuple(target)], fill="#FF0000", width=2)

            # cv2.imwrite(str(output_folder / f"image{frame:03}.png"), np_image)
            ve.write(np_image)


def load_data(
    tracking_file: Path, subsampling_factor: int, end_frame: int
) -> tuple[Overlay, nx.DiGraph]:
    """Load simple tracking ground truth

    Args:
        tracking_file (Path): Path to json file
        subsampling_factor (int): the subsampling factor that shall be applied
        end_frame (int): Cut-off frame

    Returns:
        tuple[Overlay, nx.DiGraph]: segmentation overlay and tracking graph
    """
    with open(tracking_file, encoding="utf-8") as tr_input:
        overlay, tracking_graph = parse_simple_tracking(tr_input.read())

    sub_tracking_graph = subsample_lineage(tracking_graph, subsampling_factor)

    nds_to_delete = []
    for node in sub_tracking_graph.nodes:
        if sub_tracking_graph.nodes[node]["frame"] > end_frame:
            nds_to_delete.append(node)

    sub_tracking_graph.remove_nodes_from(nds_to_delete)

    sub_nodes = set(sub_tracking_graph.nodes)

    sub_overlay = Overlay([cont for cont in overlay if cont.id in sub_nodes])

    # correct frame number
    for cont in overlay:
        cont.frame = int(np.round(cont.frame / subsampling_factor))
    for n in tracking_graph.nodes:
        tracking_graph.nodes[n]["frame"] = int(
            np.round(tracking_graph.nodes[n]["frame"] / subsampling_factor)
        )

    return sub_overlay, sub_tracking_graph



"""Utilizes the MIP wrapper library for solving assignment problems
"""

import numpy as np
from mip import BINARY, CBC, Model, maximize, minimize, xsum


class SimpleMIPSolver:
    """MIP based solver wrapper to solve the assignment prolbem"""

    def __init__(self, assignments, type="max", threads=1, solver_name=CBC):
        self.assignments = assignments

        sources, targets, costs, _ = assignments

        self.sources = sources
        self.targets = targets

        costs = np.concatenate(costs)

        self.num_assignments = len(costs)

        # costs are one-dimensional
        self.costs = np.array(costs).flatten()

        if solver_name == "auto":
            # let library choose installed optimizer (prefers gurobi)
            solver_name = ""

        m = Model(solver_name=solver_name)

        # set amount of threads
        m.threads = threads

        # Disable outputs
        m.verbose = 0

        self.m = m
        self.variables = []

        # create the variables for the assignment
        for _ in costs:
            self.variables.append(m.add_var(var_type=BINARY))

        self.variables = np.array(self.variables)

        if type == "max":
            opt_type = maximize
        else:
            opt_type = minimize

        # create the objective
        m.objective = opt_type(
            xsum(self.variables[i] * costs[i] for i in range(len(costs)))
        )

        all_sources = np.concatenate([s.flatten() for s in sources])
        valid_sources = all_sources[all_sources >= 0]
        unqiue_sources = np.unique(valid_sources)
        unique_targets = np.unique(np.concatenate([t.flatten() for t in targets]))

        # source touched exactly once
        for source in unqiue_sources:
            all_masks = []
            for ass_sources in sources:
                if ass_sources.shape[1] == 0:
                    # in case we have no sources
                    all_masks.append(np.zeros(ass_sources.shape[0], dtype=bool))
                    continue

                assignment_mask = np.isin(ass_sources, source)
                if len(assignment_mask.shape) == 2:
                    assignment_mask = np.any(assignment_mask, axis=-1)
                all_masks.append(assignment_mask)

            full_mask = np.concatenate(all_masks)

            # add a joint constraint for all the assignments that are masked true
            m += xsum(self.variables[full_mask]) == 1
            # m.addConstr(gp.quicksum(self.variables[full_mask]), GRB.EQUAL, 1)

        # target touched exactly once
        for target in unique_targets:
            all_masks = []
            for ass_targets in targets:
                if ass_targets.shape[1] == 0:
                    # in case we have no targets
                    all_masks.append(np.zeros(ass_targets.shape[0], dtype=bool))
                    continue

                assignment_mask = np.isin(ass_targets, target)
                if len(assignment_mask.shape) == 2:
                    assignment_mask = np.any(assignment_mask, axis=-1)
                all_masks.append(assignment_mask)

            full_mask = np.concatenate(all_masks)

            # add a joint constraint for all the assignments that are masked true
            m += xsum(self.variables[full_mask]) == 1
            # m.addConstr(gp.quicksum(self.variables[full_mask]), GRB.EQUAL, 1)

    def setCutOff(self, cutOff: float):
        # pylint: disable=unused-argument,no-self-use
        # self.m.Params.cutOff = cutOff
        # logging.warning("Not supported!")
        pass

    def generateSolution(self, i):

        if i != 0:
            raise NotImplementedError()

        # TODO: implement the extraction of multiple solutions

        mask = np.array([x.x >= 0.99 for x in self.variables], dtype=bool)

        # determine assignments
        used_assignments = mask  # np.array(self.assignments)[mask]

        total_costs = self.costs.dot(mask)

        return (used_assignments, total_costs)

    def generateAllSolutions(self):
        return [self.generateSolution(0)]

    def solve(self):
        """
        Search for best solution

        returns: a solution object
        """
        self.m.optimize()

        return self.generateSolution(0)

    # pylint: disable=unused-variable, unused-argument
    def populate(self, max_num_solutions, pool_gap=None):
        if max_num_solutions > 1:
            raise NotImplementedError(
                "Obtaining multiple solutions is not yet supported!"
            )

        # TODO: Implement the solution population

        # pylint: disable=pointless-string-statement
        """
        self.m.optimize()

        x: variables
        a: previous solution

        xsum(x(i) * a(i) for i in) - xsum(x(i) * (1-a(i)) for i in) <= xsum(a(i) for i in) - 1

        See: https://stackoverflow.com/questions/42591384/how-to-ask-for-second-best-solution-to-a-mip-using-jump

        """

        # self.m.optimize()
        # not yet implemented
        # logging.warning("This is not implemented")
        return [self.solve()]
        # return self.generateAllSolutions()



""" Module for assignment generation """

from __future__ import annotations

import itertools
import logging
from typing import Any

import numpy as np


def filter_targets(source_index, target_index, filters):
    """Filter targets

    Args:
        source_index (_type_): index array for the sources
        target_index (_type_): index array of SxT (source, filtered candidates)
        filters (_type_): filter functions returning boolean arrays for (source_index, target_index)

    Returns:
        tuple: [source_index, filtered target_index]
    """

    num_sources = len(source_index)
    num_targets = target_index.shape[1]

    if len(filters) == 0:
        # no filters at all
        combined_mask = np.ones((num_sources, num_targets), dtype=bool)
    else:
        # apply all filters to the source, index combinations
        stacked_masks = np.array([f(source_index, target_index) for f in filters])
        # only when all filters aggree use the
        combined_mask = np.all(stacked_masks, axis=0)

    # pylint: disable=singleton-comparison
    if np.all(combined_mask == False):
        logging.warning("Filtered all away!")

    # just keep the valid (source, target) combinations
    mask_array = np.ma.masked_array(target_index, mask=~combined_mask)
    target_index_new = mask_array.compressed().reshape((num_sources, -1))

    return source_index, target_index_new


class SimpleAssignmentGenerator:
    """
    Base class for assignment generators
    """

    def __init__(self, models: list):
        """Create an assignment generator

        Args:
            models (list): List of models to score the assignments
        """
        self.models = models

    def generate(self, tracking, sources, targets):
        raise NotImplementedError()

    def compute_scores(
        self,
        tracking: np.ndarray[Any, np.uint32],
        source_index: np.ndarray[(Any, 1), np.uint32],
        target_index: np.ndarray[(Any, Any), np.uint32],
    ) -> tuple[np.float32, np.ndarray[(Any, Any), np.float32]]:
        """Compute the scores for all assignments

        Args:
            tracking (np.ndarray[Any, np.uint32]): index based tracking representation
            source_index (np.ndarray[(Any, 1), np.uint32]) : sources for the assignemnts
            target_index (np.ndarray[(Any, Any), np.uint32]): targets for the assignment

        Raises:
            ValueError: we need models for scoring

        Returns:
            _type_: sum of scores and a list of individual scores
        """

        if len(self.models) <= 0:
            raise ValueError("You need at least one model for scoring assignments")

        scores = np.array(
            [m(tracking, source_index, target_index) for m in self.models]
        )
        assert len(scores.shape) == 2
        return np.sum(scores, axis=0), scores.T


class SimpleNewAssGenerator(SimpleAssignmentGenerator):
    """
    Generates new detection assignments
    """

    def generate(self, tracking, sources, targets) -> tuple:
        # for new detections the source is -1
        source_index = np.ones((len(targets), 1), dtype=np.int32) * -1

        # the target indices are simple the targets but with 2 dimensions
        target_index = np.zeros((len(targets), 1), dtype=np.int32)
        target_index[:, 0] = targets

        summed_scores, individual_scores = self.compute_scores(
            tracking, source_index, target_index
        )

        return (
            source_index,
            target_index,
            summed_scores,
            individual_scores,
        )


class SimpleEndTrackAssGenerator(SimpleAssignmentGenerator):
    """
    Generates new detection assignments
    """

    def generate(self, tracking, sources, targets) -> tuple:
        # for dissappearing cells copy the sources
        source_index = np.zeros((len(sources), 1), dtype=np.int32)
        source_index[:, 0] = sources

        # the target index has dim=0 (no connections)
        target_index = np.zeros((len(sources), 0), dtype=np.int32)

        # compute the scores
        summed_scores, individual_scores = self.compute_scores(
            tracking, source_index, target_index
        )

        # return all variables
        return (
            source_index,
            target_index,
            summed_scores,
            individual_scores,
        )


class SimpleContinueGenerator(SimpleAssignmentGenerator):
    """Generator for continue assignments"""

    def __init__(self, candidate_filters, models):
        """
        candidate_filters: filter possible target candidates w.r.t to a source
        models: models for scoring the assignments
        """
        super().__init__(models)
        self.candidate_filters = candidate_filters

    def generate(self, tracking, sources, targets):
        if len(sources) == 0:
            # the case for no sources (e.g. start of the tracking): no continue assignments
            return (
                np.zeros((0, 0), dtype=np.uint32),
                np.zeros((0, 0), dtype=np.uint32),
                np.array([], dtype=np.float32),
                np.array([], dtype=np.float32),
            )

        # setup index lists
        source_index = sources
        # targets are TxSx1
        target_index = np.tile(targets, (len(sources), 1))

        source_index, target_index = filter_targets(
            source_index, target_index, self.candidate_filters
        )

        source_index = np.repeat(source_index, target_index.shape[1]).reshape((-1, 1))
        target_index = target_index.reshape(-1, 1)

        assert source_index.shape == target_index.shape
        assert source_index.shape[1] == 1

        # sum up the logarithms (product in prob space)
        summed_scores, individual_scores = self.compute_scores(
            tracking, source_index, target_index
        )  # np.sum([m(tracking, source_index, target_index) for m in self.models], axis=0)

        assert summed_scores.shape[0] == source_index.shape[0]

        return source_index, target_index, summed_scores, individual_scores


class SimpleSplitGenerator(SimpleAssignmentGenerator):
    """Generator for division assignments"""

    def __init__(self, candidate_filters, pair_filter, models):
        super().__init__(models)
        self.candidate_filters = candidate_filters
        self.pair_filter = pair_filter

    def generate(self, tracking, sources, targets):
        if len(sources) == 0:
            return (
                np.zeros((0, 0), dtype=np.uint32),
                np.zeros((0, 0), dtype=np.uint32),
                np.array([], dtype=np.float32),
                np.array([], dtype=np.float32),
            )

        source_index = sources
        # create target matrix
        target_index = np.tile(np.array(targets), (len(sources), 1))

        # apply filters for target indices
        source_index, target_index = filter_targets(
            source_index, target_index, self.candidate_filters
        )

        combinations = np.array(
            [list(itertools.combinations(targets, 2)) for targets in target_index]
        )

        source_index = np.repeat(sources, (combinations.shape[1],)).reshape((-1, 1))
        target_index = combinations.reshape((-1, 2))

        # filter lists to only appropriate assignments
        mask = self.pair_filter(source_index, target_index)

        source_index = source_index[mask]
        target_index = target_index[mask]

        # check whether we still have opportunities
        if len(source_index) == 0 or len(target_index) == 0:
            # we have no split opportunities
            print("no splits available!")
            return (
                np.zeros((0, 0), dtype=np.uint32),
                np.zeros((0, 0), dtype=np.uint32),
                np.array([], dtype=np.float32),
                np.array([], dtype=np.float32),
            )

        # print(source_index, target_index)
        assert source_index.shape[0] == target_index.shape[0]
        assert source_index.shape[1] == 1
        assert target_index.shape[1] == 2

        # sum up the logarithms (product in prob space)
        summed_scores, individual_scores = self.compute_scores(
            tracking, source_index, target_index
        )  # np.sum([m(tracking, source_index, target_index) for m in self.models], axis=0)

        assert source_index.shape[0] == summed_scores.shape[0]
        assert len(summed_scores.shape) == 1

        return source_index, target_index, summed_scores, individual_scores



"""Utilize the gurobi interface for solving the assignment problem"""

import logging

import gurobipy as gp
import numpy as np
from gurobipy import GRB


class SimpleGurobiSolver:
    """Gurobi solver wrapper for the assignment problem"""

    def __init__(self, assignments, type="max", threads=1):
        self.assignments = assignments

        sources, targets, costs, _ = assignments

        self.sources = sources
        self.targets = targets

        costs = np.concatenate(costs)

        self.num_assignments = len(costs)

        # costs are one-dimensional
        costs = np.array(costs).flatten()

        gp.setParam(GRB.Param.Threads, threads)
        gp.setParam(GRB.Param.OutputFlag, 0)

        m = gp.Model("expansion")

        self.m = m
        self.variables = []

        # create the variables for the assignment
        for _ in costs:
            self.variables.append(m.addVar(vtype=GRB.BINARY))

        self.variables = np.array(self.variables)

        if type == "max":
            grb_type = GRB.MAXIMIZE
        else:
            grb_type = GRB.MINIMIZE

        # create the objective
        expr = gp.LinExpr()
        for i, _ in enumerate(costs):
            expr.add(self.variables[i], costs[i])
        m.setObjective(expr, grb_type)

        all_sources = np.concatenate([s.flatten() for s in sources])
        valid_sources = all_sources[all_sources >= 0]
        unqiue_sources = np.unique(valid_sources)
        unique_targets = np.unique(np.concatenate([t.flatten() for t in targets]))

        # source touched exactly once
        for source in unqiue_sources:
            all_masks = []
            for ass_sources in sources:
                if ass_sources.shape[1] == 0:
                    # in case we have no sources
                    all_masks.append(np.zeros(ass_sources.shape[0], dtype=bool))
                    continue

                assignment_mask = np.isin(ass_sources, source)
                if len(assignment_mask.shape) == 2:
                    assignment_mask = np.any(assignment_mask, axis=-1)
                all_masks.append(assignment_mask)

            full_mask = np.concatenate(all_masks)

            # add a joint constraint for all the assignments that are masked true
            m.addConstr(gp.quicksum(self.variables[full_mask]), GRB.EQUAL, 1)

        # target touched exactly once
        for target in unique_targets:
            all_masks = []
            for ass_targets in targets:
                if ass_targets.shape[1] == 0:
                    # in case we have no targets
                    all_masks.append(np.zeros(ass_targets.shape[0], dtype=bool))
                    continue

                assignment_mask = np.isin(ass_targets, target)
                if len(assignment_mask.shape) == 2:
                    assignment_mask = np.any(assignment_mask, axis=-1)
                all_masks.append(assignment_mask)

            full_mask = np.concatenate(all_masks)

            # add a joint constraint for all the assignments that are masked true
            m.addConstr(gp.quicksum(self.variables[full_mask]), GRB.EQUAL, 1)

    def setCutOff(self, cutOff: float):
        self.m.Params.cutOff = cutOff

    def generateSolution(self, i):
        # check whether enough solutions are available
        if i >= self.m.SolCount:
            raise ValueError(
                f"You want to get solution {i} but only {self.m.SolCount} are available!"
            )

        # select the corresponding solution
        self.m.setParam(GRB.Param.SolutionNumber, i)

        # determine assignments
        used_assignments = [
            self.variables[i].Xn > 0.9 for i, _ in enumerate(self.variables)
        ]

        total_costs = self.m.PoolObjVal

        return (used_assignments, total_costs)

    def generateAllSolutions(self):
        return [self.generateSolution(i) for i in range(self.m.SolCount)]

    def solve(self):
        """
        Search for best solution

        returns: a solution object
        """
        self.m.setParam(GRB.Param.PoolSearchMode, 0)

        self.m.optimize()

        return self.generateSolution(0)

    def populate(self, max_num_solutions, pool_gap=None):
        # Limit how many solutions to collect
        self.m.setParam(GRB.Param.PoolSolutions, max_num_solutions)
        # Limit the search space by setting a gap for the worst possible solution
        # that will be accepted
        if pool_gap:
            self.m.setParam(GRB.Param.PoolGap, pool_gap)
        # do a systematic search for the k-best solutions
        self.m.setParam(GRB.Param.PoolSearchMode, 2)

        # Optimize model
        self.m.optimize()

        nSolutions = self.m.SolCount

        logging.info(f"Number of solutions: {nSolutions}")

        if self.m.Status == GRB.TIME_LIMIT:
            logging.warning("Time limit reached! Solutions found %d", nSolutions)

        try:
            return [self.generateSolution(i) for i in range(nSolutions)]
        except AttributeError:
            logging.warning(
                "Attribute error while generating solutions! No solutions found!"
            )
            return []



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup

with open("README.md", encoding="utf-8") as readme_file:
    readme = readme_file.read()

with open("requirements.txt", encoding="utf-8") as req_file:
    requirements = req_file.read().splitlines()

test_requirements = ["pytest>=3", "wget"]

setup(
    author="Johannes Seiffarth",
    author_email="j.seiffarth@fz-juelich.de",
    python_requires=">=3.6",
    classifiers=[],
    description="Uncertainty-Aware Tracking Framework.",
    entry_points={
        "console_scripts": [],
    },
    install_requires=requirements,
    extras_require={"gurobi": ["gurobipy>=5.9.3"], "test": test_requirements},
    license="MIT license",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="uatrack",
    name="uatrack",
    packages=find_packages(include=["uatrack", "uatrack.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/JuBiotech/PyUAT",
    version="0.0.5",
    zip_safe=False,
)



""" Test tracking models
"""
import time
import unittest
from pathlib import Path

import wget

from uatrack.config import setup_assignment_generators
from uatrack.core import simpleTracking
from uatrack.utils import extract_single_cell_information, load_data, save_tracking


class TestTracking(unittest.TestCase):
    """Test contour functionality"""

    def setUp(self):
        self.segmentation = Path("filtered_0.json")
        self.image_stack = Path("00_stack.tif")

        # Download data if needed
        if not self.segmentation.exists():
            self.segmentation = wget.download(
                "https://fz-juelich.sciebo.de/s/5vBB6tW8c2DpaU3/download"
            )

        if not self.image_stack.exists():
            self.image_stack = wget.download(
                "https://fz-juelich.sciebo.de/s/Xge7fj56QM5ev7q/download"
            )

    def tracking(self, model="NN", end_frame=100):
        tracking_file = self.segmentation

        subsampling_factor = 1

        overlay, _ = load_data(
            tracking_file, subsampling_factor=subsampling_factor, end_frame=end_frame
        )

        output_file = "simpleTracking.json.gz"

        # extract arguments
        num_particles = 1  # args.nb_particles
        num_cores = 1  # args.nb_cpus
        max_num_hypotheses = 1  # args.nb_max_hypotheses
        cutOff = -1  # -10  # args.cutOff
        max_num_solutions = 1  # 10  # args.sol_pool_size

        print("Extract single-cell information...")
        df, all_detections = extract_single_cell_information(overlay)

        print("Setup assignment generators...")
        assignment_generators = setup_assignment_generators(
            df, subsampling_factor=subsampling_factor, model=model
        )

        print("Perform tracking...")
        # start tracking
        start = time.time()
        res = simpleTracking(
            df,
            assignment_generators,
            num_particles,
            num_cores=num_cores,
            max_num_hypotheses=max_num_hypotheses,
            cutOff=cutOff,
            max_num_solutions=max_num_solutions,
            mip_method="CBC",  # use CBC as gurobi is not installed in colab (there are gurobi colab examples. Thus, it should be possible to use gurobi)
        )
        end = time.time()

        print("time for tracking", end - start)

        save_tracking(res[0], all_detections, output_file)

    def test__nn_tracking(self):
        self.tracking(model="NN")

    def test__fo_tracking(self):
        self.tracking(model="FO")

    def test__fo_g_tracking(self):
        self.tracking(model="FO+G")

    def test__fo_o_tracking(self):
        self.tracking(model="FO+O")

    def test__fo_dd_tracking(self):
        self.tracking(model="FO+DD")

    def test__fo_g_o_dd_tracking(self):
        self.tracking(model="FO+G+O+DD")


if __name__ == "__main__":
    unittest.main()



"""Functionality for extracting information and scoring it according to probabilistic models"""

from __future__ import annotations

import numpy as np
from tensor_walks.impl_np import NP_Impl

backend = NP_Impl()


class ConstantModel:
    """
    model for constant probability
    """

    def __init__(self, const_log_value):
        self.const_log_value = const_log_value

    def __call__(self, tracking, source_index, target_index):
        return np.ones(source_index.shape[0]) * self.const_log_value


class ModelExecutor:
    """General model executor class
    It contains a quantity_computer and a (statistical) model.
    The quantity_computer extracts the information from suggested assignments (e.g. growth).
    The model computes a score for the extracted values."""

    def __init__(self, quantity_computer, model, df, name=None):
        self.quantity_computer = quantity_computer
        self.model = model
        self.df = df
        self.name = name

    def __call__(self, tracking, source_index, target_index):
        values = self.quantity_computer(tracking, source_index, target_index, self.df)

        return np.clip(self.model(values), -40, 0)

    def compute_quantities(self, tracking, source_index, target_index):
        values = self.quantity_computer(tracking, source_index, target_index, self.df)

        return values

    def __repr__(self):
        return f"ModelExecutor({self.name})"


# ---------------------------
# --- Quantity extractors ---
# ---------------------------


def area_growth_computer(tracking, source_index, target_index, df):
    """
    Computes the joint area growth between source and targets for all assignments
    """
    # pylint: disable=unused-argument

    assert source_index.shape[1] == 1 and len(source_index.shape) == 2
    assert len(target_index.shape) == 2

    # fetch area
    source_areas = df["area"][
        source_index.flatten()
    ]  # df['area'].to_numpy()[source_index.flatten()]
    target_areas = df["area"][target_index]  # .to_numpy()[target_index]

    # sum up the target areas
    target_joint_area = np.sum(target_areas, axis=-1)

    assert target_joint_area.shape == source_areas.shape

    # compute growth
    return np.divide(target_joint_area, source_areas)


def predict_movement(
    tracking,
    source_index,
    df,
    history: int,
    alpha=None,
    stop_at_split=True,
):
    """
    Predict the movement of a cell based on its past movements (in history timesteps). If no history is available assume no movement.

    Returns the distance between the prediction and the true positions

    """
    # fetch centroids
    all_centroids = df[
        "centroid"
    ]  # np.array(df['centroid'].to_list(), dtype=np.float32)

    # compute unique array
    unique_sources, inverse_indexing = np.unique(source_index, return_inverse=True)

    # perform walks from unique sources
    parents = backend.compute_walks_upward(
        tracking.parents, unique_sources, history, stop_at_split=stop_at_split
    )

    # get the centroids of the walks
    # walk_centroids = backend.compute_centroids(parents, all_centroids)

    # compute the average movement along the walks
    avg_movement = backend.compute_avg_property_along_walk(
        parents, all_centroids, exp_moving_average=alpha
    )

    # predict the next position of the sources
    pred_position = all_centroids[unique_sources] + avg_movement

    # rearrange predicted positions to assignment shape
    source_predictions = pred_position[inverse_indexing, :]

    # for all the sources without history we assume no movement
    pred_mask = np.all(source_predictions.mask, axis=1)
    source_predictions[pred_mask, :] = all_centroids[
        source_index[pred_mask].flatten(), :
    ]

    return source_predictions


def predict_property(
    tracking,
    source_index,
    df,
    history: int,
    alpha=None,
    stop_at_split=True,
    property_name="area",
):
    """
    Predicts the development of a single-cell property based on its past movements (in history timesteps). If no history is available assume no development.

    Returns the predicted property development
    """
    # fetch centroids
    all_properties = df[
        property_name
    ]  # np.array(df['centroid'].to_list(), dtype=np.float32)

    # auto compute property dimension
    prop_dim = 1
    if len(all_properties.shape) == 2:
        prop_dim = all_properties.shape[-1]

    # compute unique array
    unique_sources, inverse_indexing = np.unique(source_index, return_inverse=True)

    # perform walks from unique sources
    parents = backend.compute_walks_upward(
        tracking.parents, unique_sources, history, stop_at_split=stop_at_split
    )

    # get the centroids of the walks
    # walk_centroids = backend.compute_centroids(parents, all_centroids)

    # compute the average movement along the walks
    avg_development = backend.compute_avg_property_along_walk(
        parents, all_properties, exp_moving_average=alpha, distance_dims=prop_dim
    )

    # predict the next position of the sources
    pred_position = all_properties[unique_sources] + avg_development

    # rearrange predicted positions to assignment shape
    source_predictions = pred_position[inverse_indexing, ...]

    # for all the sources without history we assume no movement
    pred_mask = np.all(source_predictions.mask, axis=-1)
    if np.any(pred_mask):
        source_predictions[pred_mask] = all_properties[
            source_index[pred_mask].flatten()
        ]

    return source_predictions


def first_order_growth(
    tracking,
    source_index,
    target_index,
    df,
    history: int,
    alpha=None,
    stop_at_split=True,
):
    # extract all areas
    all_areas = df["area"]

    # assert target_index.shape[1] == 1

    # predict area development
    source_predictions = predict_property(
        tracking,
        source_index=source_index,
        df=df,
        history=history,
        stop_at_split=stop_at_split,
        property_name="area",
        alpha=alpha,
    )
    # source_predictions = np.repeat(source_predictions, target_index.shape[1], axis=0)

    target_areas = all_areas[target_index]  # .to_numpy()[target_index]

    # sum up the target areas
    target_joint_area = np.sum(target_areas, axis=-1)

    # print("Dimensions:")
    # print(source_predictions.shape, target_joint_area.shape)

    # compute area difference between sources and targets
    assignment_distances = source_predictions - target_joint_area

    # print(assignment_distances.shape)

    # return the distances
    return assignment_distances


def distance_to_pred_computer(
    tracking,
    source_index,
    target_index,
    df,
    history: int,
    alpha=None,
    stop_at_split=True,
):
    """
    Predict the movement of a cell based on its past movements (in history timesteps). If no history is available assume no movement.

    Returns the distance between the prediction and the true positions

    """
    # fetch centroids
    all_centroids = df["centroid"]

    # predict position by movement estimation
    source_predictions = predict_property(
        tracking,
        source_index=source_index,
        df=df,
        history=history,
        stop_at_split=stop_at_split,
        property_name="centroid",
        alpha=alpha,
    )

    # extract target positions (positions of cell detections in the next frame)
    target_positions = all_centroids[target_index]

    # compute distances between predicitions and target positions
    # we introduce a new dimensions because there can be multiple target positions (e.g. two for division into two children)
    assignment_distances = np.linalg.norm(
        source_predictions[:, None, :] - target_positions, axis=-1
    )

    # Sum multiple distances distances
    assignment_distances = np.sum(assignment_distances, axis=1)

    # return the distances
    return assignment_distances


def distance_to_pred_masked(tracking, source_index, target_index, df, num_steps=3):
    walks = backend.compute_walks_upward(
        tracking.parents,
        source_index.reshape(-1),
        num_steps=num_steps,
        stop_at_split=True,
    )

    all_positions = df["centroid"]

    avg_movement = backend.compute_avg_property_along_walk(walks, all_positions)

    pred_position = all_positions[source_index.reshape(-1)] + avg_movement

    target_position = all_positions[target_index.reshape(-1)]

    pred_distance = np.abs(np.linalg.norm(pred_position - target_position, axis=1))

    invalid = np.any(pred_position.mask)

    return np.ma.array(pred_distance, invalid)


def compute_dist_to_last_split(tracking, walk_matrix, no_split_behavior="mask"):
    """
    tracking: The current parent lookup table
    walk_matrix: NxM walk matrix for N sources and at max M steps
    no_split_behavior: if "mask" return masked array, otherwise invalid entries are filled with distance to root

    return distance array (N,)
    """
    # now we have to measure distance to last split
    numChildren = backend.compute_num_children(tracking)

    # the number of children for every parent
    data = np.ma.masked_array(
        numChildren[walk_matrix[:, 1:]], mask=~(walk_matrix[:, 1:] >= 0)
    )
    # np.argmax returns first occurence index of True, data > 1 means node did split
    all_dist = np.argmax(data > 1, axis=1)

    if no_split_behavior == "mask":
        all_dist = np.ma.masked_array(all_dist, mask=~(np.any(data > 1, axis=1)))
    else:
        invalid_entris = np.all(data <= 1, axis=1)

        # for all invalid entries compute the number of steps to root
        all_dist[invalid_entris] = np.sum((walk_matrix >= 0)[invalid_entris], axis=1)

    return all_dist


def split_dist_computer(
    tracking, source_index, target_index, df, max_time_consideration=100
):
    """
    Computes the number of timesteps to the last division event

    max_time_consideration: Maximum steps of time to reach into the past to find the parents

    returns the distance to the last split or -1 if the last split could not be found
    """
    # pylint: disable=unused-argument

    # fetch the unique source ideas
    unique_sources = np.unique(source_index)

    # initialize the distances to (-1): -1 means
    distances = np.ones(np.max(unique_sources) + 1, dtype=np.int32) * -1

    parents = backend.compute_walks_upward(
        tracking.parents, unique_sources, max_time_consideration
    )

    masked_distances = compute_dist_to_last_split(tracking.parents, parents)

    distances[unique_sources] = masked_distances.data

    distances[unique_sources[masked_distances.mask]] = -1

    return distances[source_index.flatten()]



""" Test tracking models
"""
import unittest

import numpy as np

from uatrack.models import predict_movement, predict_property


class FakeTracking:
    """Make a fake tracking"""

    def __init__(self, tracking_data):
        self.parents = tracking_data


class TestModels(unittest.TestCase):
    """Test contour functionality"""

    @staticmethod
    def test_movement_predict():
        tracking = FakeTracking(np.array([-1, 0, 0, 2]))
        data = {
            "centroid": np.array(
                [
                    [0, 0],
                    [2, 0],
                    [0, 2],
                    [0, 0],
                ]
            )
        }

        # predict movement for one step
        pmovement = predict_movement(
            tracking,
            np.array([1, 2], int).reshape((-1, 1)),
            data,
            1,
            None,
            stop_at_split=False,
        )
        np.testing.assert_almost_equal(pmovement, np.array([[4, 0], [0, 4]]))

        # predict movement for one step
        pmovement = predict_movement(
            tracking,
            np.array([1, 2], int).reshape((-1, 1)),
            data,
            1,
            None,
            stop_at_split=True,
        )
        np.testing.assert_almost_equal(pmovement, np.array([[2, 0], [0, 2]]))

    @staticmethod
    def test_movement_predict_general():
        tracking = FakeTracking(np.array([-1, 0, 0, 2]))
        data = {
            "centroid": np.array(
                [
                    [0, 0],
                    [2, 0],
                    [0, 2],
                    [0, 0],
                ]
            )
        }

        # predict movement for one step
        pmovement = predict_property(
            tracking,
            np.array([1, 2], int).reshape((-1, 1)),
            data,
            1,
            None,
            stop_at_split=False,
            property_name="centroid",
        )
        np.testing.assert_almost_equal(pmovement, np.array([[4, 0], [0, 4]]))

        # predict movement for one step
        pmovement = predict_property(
            tracking,
            np.array([1, 2], int).reshape((-1, 1)),
            data,
            1,
            None,
            stop_at_split=True,
            property_name="centroid",
        )
        np.testing.assert_almost_equal(pmovement, np.array([[2, 0], [0, 2]]))

    @staticmethod
    def test_growth_predict():
        tracking = FakeTracking(np.array([-1, 0, 1, 2]))
        data = {"area": np.array([2, 4, 8, 12])}  # 0  # 1  # 2  # 3

        # predict movement for one step
        pmovement = predict_property(
            tracking,
            np.array([1, 2], int).reshape((-1, 1)),
            data,
            1,
            None,
            property_name="area",
        )
        np.testing.assert_almost_equal(pmovement, np.array([6, 12]))

        pmovement = predict_property(
            tracking,
            np.array([0, 1, 2], int).reshape((-1, 1)),
            data,
            1,
            None,
            property_name="area",
        )
        np.testing.assert_almost_equal(pmovement, np.array([2, 6, 12]))
        # predict movement for one step
        # pmovement = predict_movement(tracking, np.array([1,2], int).reshape((-1, 1)), data, 1, None, stop_at_split=True)
        # np.testing.assert_almost_equal(pmovement, np.array([[2, 0], [0, 2]]))


if __name__ == "__main__":
    unittest.main()
